# Copyright (C) 2007, 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Python implementations of Dirstate Helper functions."""

import binascii
import os
import struct

# We cannot import the dirstate module, because it loads this module
# All we really need is the IN_MEMORY_MODIFIED constant
from .dirstate import DirState, DirstateCorrupt


def pack_stat(st, _b64=binascii.b2a_base64, _pack=struct.Struct('>6L').pack):
    """Convert stat values into a packed representation

    Not all of the fields from the stat included are strictly needed, and by
    just encoding the mtime and mode a slight speed increase could be gained.
    However, using the pyrex version instead is a bigger win.
    """
    # base64 encoding always adds a final newline, so strip it off
    return _b64(_pack(st.st_size & 0xFFFFFFFF, int(st.st_mtime) & 0xFFFFFFFF,
                      int(st.st_ctime) & 0xFFFFFFFF, st.st_dev & 0xFFFFFFFF,
                      st.st_ino & 0xFFFFFFFF, st.st_mode))[:-1]


def _unpack_stat(packed_stat):
    """Turn a packed_stat back into the stat fields.

    This is meant as a debugging tool, should not be used in real code.
    """
    (st_size, st_mtime, st_ctime, st_dev, st_ino,
     st_mode) = struct.unpack('>6L', binascii.a2b_base64(packed_stat))
    return dict(st_size=st_size, st_mtime=st_mtime, st_ctime=st_ctime,
                st_dev=st_dev, st_ino=st_ino, st_mode=st_mode)


def _bisect_path_left(paths, path):
    """Return the index where to insert path into paths.

    This uses the dirblock sorting. So all children in a directory come before
    the children of children. For example::

        a/
          b/
            c
          d/
            e
          b-c
          d-e
        a-a
        a=c

    Will be sorted as::

        a
        a-a
        a=c
        a/b
        a/b-c
        a/d
        a/d-e
        a/b/c
        a/d/e

    :param paths: A list of paths to search through
    :param path: A single path to insert
    :return: An offset where 'path' can be inserted.
    :seealso: bisect.bisect_left
    """
    hi = len(paths)
    lo = 0
    while lo < hi:
        mid = (lo + hi) // 2
        # Grab the dirname for the current dirblock
        cur = paths[mid]
        if _lt_path_by_dirblock(cur, path):
            lo = mid + 1
        else:
            hi = mid
    return lo


def _bisect_path_right(paths, path):
    """Return the index where to insert path into paths.

    This uses a path-wise comparison so we get::
        a
        a-b
        a=b
        a/b
    Rather than::
        a
        a-b
        a/b
        a=b
    :param paths: A list of paths to search through
    :param path: A single path to insert
    :return: An offset where 'path' can be inserted.
    :seealso: bisect.bisect_right
    """
    hi = len(paths)
    lo = 0
    while lo < hi:
        mid = (lo + hi) // 2
        # Grab the dirname for the current dirblock
        cur = paths[mid]
        if _lt_path_by_dirblock(path, cur):
            hi = mid
        else:
            lo = mid + 1
    return lo


def bisect_dirblock(dirblocks, dirname, lo=0, hi=None, cache={}):
    """Return the index where to insert dirname into the dirblocks.

    The return value idx is such that all directories blocks in dirblock[:idx]
    have names < dirname, and all blocks in dirblock[idx:] have names >=
    dirname.

    Optional args lo (default 0) and hi (default len(dirblocks)) bound the
    slice of a to be searched.
    """
    if hi is None:
        hi = len(dirblocks)
    try:
        dirname_split = cache[dirname]
    except KeyError:
        dirname_split = dirname.split(b'/')
        cache[dirname] = dirname_split
    while lo < hi:
        mid = (lo + hi) // 2
        # Grab the dirname for the current dirblock
        cur = dirblocks[mid][0]
        try:
            cur_split = cache[cur]
        except KeyError:
            cur_split = cur.split(b'/')
            cache[cur] = cur_split
        if cur_split < dirname_split:
            lo = mid + 1
        else:
            hi = mid
    return lo


def lt_by_dirs(path1, path2):
    """Compare two paths directory by directory.

    This is equivalent to doing::

       operator.lt(path1.split('/'), path2.split('/'))

    The idea is that you should compare path components separately. This
    differs from plain ``path1 < path2`` for paths like ``'a-b'`` and ``a/b``.
    "a-b" comes after "a" but would come before "a/b" lexically.

    :param path1: first path
    :param path2: second path
    :return: True if path1 comes first, otherwise False
    """
    if not isinstance(path1, bytes):
        raise TypeError("'path1' must be a byte string, not %s: %r"
                        % (type(path1), path1))
    if not isinstance(path2, bytes):
        raise TypeError("'path2' must be a byte string, not %s: %r"
                        % (type(path2), path2))
    return path1.split(b'/') < path2.split(b'/')


def _lt_path_by_dirblock(path1, path2):
    """Compare two paths based on what directory they are in.

    This generates a sort order, such that all children of a directory are
    sorted together, and grandchildren are in the same order as the
    children appear. But all grandchildren come after all children.

    :param path1: first path
    :param path2: the second path
    :return: True if path1 comes first, otherwise False
    """
    if not isinstance(path1, bytes):
        raise TypeError("'path1' must be a plain string, not %s: %r"
                        % (type(path1), path1))
    if not isinstance(path2, bytes):
        raise TypeError("'path2' must be a plain string, not %s: %r"
                        % (type(path2), path2))
    dirname1, basename1 = os.path.split(path1)
    key1 = (dirname1.split(b'/'), basename1)
    dirname2, basename2 = os.path.split(path2)
    key2 = (dirname2.split(b'/'), basename2)
    return key1 < key2


def _read_dirblocks(state):
    """Read in the dirblocks for the given DirState object.

    This is tightly bound to the DirState internal representation. It should be
    thought of as a member function, which is only separated out so that we can
    re-write it in pyrex.

    :param state: A DirState object.
    :return: None
    """
    state._state_file.seek(state._end_of_header)
    text = state._state_file.read()
    # TODO: check the crc checksums. crc_measured = zlib.crc32(text)

    fields = text.split(b'\0')
    # Remove the last blank entry
    trailing = fields.pop()
    if trailing != b'':
        raise DirstateCorrupt(state,
                              'trailing garbage: {!r}'.format(trailing))
    # consider turning fields into a tuple.

    # skip the first field which is the trailing null from the header.
    cur = 1
    # Each line now has an extra '\n' field which is not used
    # so we just skip over it
    # entry size:
    #  3 fields for the key
    #  + number of fields per tree_data (5) * tree count
    #  + newline
    num_present_parents = state._num_present_parents()
    tree_count = 1 + num_present_parents
    entry_size = state._fields_per_entry()
    expected_field_count = entry_size * state._num_entries
    field_count = len(fields)
    # this checks our adjustment, and also catches file too short.
    if field_count - cur != expected_field_count:
        raise DirstateCorrupt(state,
                              'field count incorrect %s != %s, entry_size=%s, '
                              'num_entries=%s fields=%r' % (
                                  field_count - cur, expected_field_count, entry_size,
                                  state._num_entries, fields))

    if num_present_parents == 1:
        # Bind external functions to local names
        _int = int
        # We access all fields in order, so we can just iterate over
        # them. Grab an straight iterator over the fields. (We use an
        # iterator because we don't want to do a lot of additions, nor
        # do we want to do a lot of slicing)
        _iter = iter(fields)
        # Get a local reference to the compatible next method
        next = getattr(_iter, '__next__', None)
        if next is None:
            next = _iter.next
        # Move the iterator to the current position
        for x in range(cur):
            next()
        # The two blocks here are deliberate: the root block and the
        # contents-of-root block.
        state._dirblocks = [(b'', []), (b'', [])]
        current_block = state._dirblocks[0][1]
        current_dirname = b''
        append_entry = current_block.append
        for count in range(state._num_entries):
            dirname = next()
            name = next()
            file_id = next()
            if dirname != current_dirname:
                # new block - different dirname
                current_block = []
                current_dirname = dirname
                state._dirblocks.append((current_dirname, current_block))
                append_entry = current_block.append
            # we know current_dirname == dirname, so re-use it to avoid
            # creating new strings
            entry = ((current_dirname, name, file_id),
                     [(  # Current Tree
                         next(),                # minikind
                         next(),                # fingerprint
                         _int(next()),          # size
                         next() == b'y',        # executable
                         next(),                # packed_stat or revision_id
                     ),
                (  # Parent 1
                         next(),                # minikind
                         next(),                # fingerprint
                         _int(next()),          # size
                         next() == b'y',        # executable
                         next(),                # packed_stat or revision_id
                     ),
                ])
            trailing = next()
            if trailing != b'\n':
                raise ValueError("trailing garbage in dirstate: %r" % trailing)
            # append the entry to the current block
            append_entry(entry)
        state._split_root_dirblock_into_contents()
    else:
        fields_to_entry = state._get_fields_to_entry()
        entries = [fields_to_entry(fields[pos:pos + entry_size])
                   for pos in range(cur, field_count, entry_size)]
        state._entries_to_current_state(entries)
    # To convert from format 2  => format 3
    # state._dirblocks = sorted(state._dirblocks,
    #                          key=lambda blk:blk[0].split('/'))
    # To convert from format 3 => format 2
    # state._dirblocks = sorted(state._dirblocks)
    state._dirblock_state = DirState.IN_MEMORY_UNMODIFIED



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Test 'brz init'"""

import os
import re

from breezy import branch as _mod_branch
from breezy import config as _mod_config
from breezy import osutils, urlutils
from breezy.bzr.bzrdir import BzrDirMetaFormat1
from breezy.tests import TestCaseWithTransport, TestSkipped
from breezy.tests.test_sftp_transport import TestCaseWithSFTPServer
from breezy.workingtree import WorkingTree


class TestInit(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self._default_label = '2a'

    def test_init_with_format(self):
        # Verify brz init --format constructs something plausible
        t = self.get_transport()
        self.run_bzr('init --format default')
        self.assertIsDirectory('.bzr', t)
        self.assertIsDirectory('.bzr/checkout', t)
        self.assertIsDirectory('.bzr/checkout/lock', t)

    def test_init_format_2a(self):
        """Smoke test for constructing a format 2a repository."""
        out, err = self.run_bzr('init --format=2a')
        self.assertEqual("""Created a standalone tree (format: 2a)\n""",
                         out)
        self.assertEqual('', err)

    def test_init_format_bzr(self):
        """Smoke test for constructing a format with the 'bzr' alias."""
        out, err = self.run_bzr('init --format=bzr')
        self.assertEqual(
            "Created a standalone tree (format: %s)\n" % self._default_label,
            out)
        self.assertEqual('', err)

    def test_init_colocated(self):
        """Smoke test for constructing a colocated branch."""
        out, err = self.run_bzr(
            'init --format=development-colo file:,branch=abranch')
        self.assertEqual("""Created a standalone tree (format: development-colo)\n""",
                         out)
        self.assertEqual('', err)
        out, err = self.run_bzr('branches')
        self.assertEqual("  abranch\n", out)
        self.assertEqual('', err)

    def test_init_at_repository_root(self):
        # brz init at the root of a repository should create a branch
        # and working tree even when creation of working trees is disabled.
        t = self.get_transport()
        t.mkdir('repo')
        format = BzrDirMetaFormat1()
        newdir = format.initialize(t.abspath('repo'))
        repo = newdir.create_repository(shared=True)
        repo.set_make_working_trees(False)
        out, err = self.run_bzr('init repo')
        self.assertEqual("""Created a repository tree (format: {})
Using shared repository: {}
""".format(self._default_label, urlutils.local_path_from_url(
            repo.controldir.root_transport.external_url())), out)
        cwd = osutils.getcwd()
        self.assertEndsWith(out, cwd + '/repo/\n')
        self.assertEqual('', err)
        newdir.open_branch()
        newdir.open_workingtree()

    def test_init_branch(self):
        out, err = self.run_bzr('init')
        self.assertEqual("Created a standalone tree (format: {})\n".format(
            self._default_label), out)
        self.assertEqual('', err)

        # Can it handle subdirectories of branches too ?
        out, err = self.run_bzr('init subdir1')
        self.assertEqual("Created a standalone tree (format: {})\n".format(
            self._default_label), out)
        self.assertEqual('', err)
        WorkingTree.open('subdir1')

        self.run_bzr_error(['Parent directory of subdir2/nothere does not exist'],
                           'init subdir2/nothere')
        out, err = self.run_bzr('init subdir2/nothere', retcode=3)
        self.assertEqual('', out)

        os.mkdir('subdir2')
        out, err = self.run_bzr('init subdir2')
        self.assertEqual("Created a standalone tree (format: {})\n".format(
            self._default_label), out)
        self.assertEqual('', err)
        # init an existing branch.
        out, err = self.run_bzr('init subdir2', retcode=3)
        self.assertEqual('', out)
        self.assertTrue(err.startswith('brz: ERROR: Already a branch:'))

    def test_init_branch_quiet(self):
        out, err = self.run_bzr('init -q')
        self.assertEqual('', out)
        self.assertEqual('', err)

    def test_init_existing_branch(self):
        self.run_bzr('init')
        out, err = self.run_bzr('init', retcode=3)
        self.assertContainsRe(err, 'Already a branch')
        # don't suggest making a checkout, there's already a working tree
        self.assertFalse(re.search(r'checkout', err))

    def test_init_existing_without_workingtree(self):
        # make a repository
        repo = self.make_repository('.', shared=True)
        repo.set_make_working_trees(False)
        # make a branch; by default without a working tree
        self.run_bzr('init subdir')
        # fail
        out, err = self.run_bzr('init subdir', retcode=3)
        # suggests using checkout
        self.assertContainsRe(err,
                              'ontains a branch.*but no working tree.*checkout')

    def test_no_defaults(self):
        """Init creates no default ignore rules."""
        self.run_bzr('init')
        self.assertFalse(os.path.exists('.bzrignore'))

    def test_init_unicode(self):
        # Make sure getcwd can handle unicode filenames
        try:
            os.mkdir('mu-\xb5')
        except UnicodeError:
            raise TestSkipped("Unable to create Unicode filename")
        # try to init unicode dir
        self.run_bzr(['init', '-q', 'mu-\xb5'])

    def create_simple_tree(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/a'])
        tree.add(['a'], ids=[b'a-id'])
        tree.commit('one', rev_id=b'r1')
        return tree

    def test_init_create_prefix(self):
        """'brz init --create-prefix; will create leading directories."""
        tree = self.create_simple_tree()

        self.run_bzr_error(['Parent directory of ../new/tree does not exist'],
                           'init ../new/tree', working_dir='tree')
        self.run_bzr('init ../new/tree --create-prefix', working_dir='tree')
        self.assertPathExists('new/tree/.bzr')

    def test_init_default_format_option(self):
        """brz init should read default format from option default_format"""
        g_store = _mod_config.GlobalStore()
        g_store._load_from_string(b'''
[DEFAULT]
default_format = 1.9
''')
        g_store.save()
        out, err = self.run_brz_subprocess('init')
        self.assertContainsRe(out, b'1.9')

    def test_init_no_tree(self):
        """'brz init --no-tree' creates a branch with no working tree."""
        out, err = self.run_bzr('init --no-tree')
        self.assertStartsWith(out, 'Created a standalone branch')


class TestSFTPInit(TestCaseWithSFTPServer):

    def test_init(self):
        # init on a remote url should succeed.
        out, err = self.run_bzr(['init', '--format=pack-0.92', self.get_url()])
        self.assertEqual(out,
                         """Created a standalone branch (format: pack-0.92)\n""")

    def test_init_existing_branch(self):
        # when there is already a branch present, make mention
        self.make_branch('.')

        # rely on SFTPServer get_url() pointing at '.'
        out, err = self.run_bzr_error(['Already a branch'],
                                      ['init', self.get_url()])

        # make sure using 'brz checkout' is not suggested
        # for remote locations missing a working tree
        self.assertFalse(re.search(r'use brz checkout', err))

    def test_init_existing_branch_with_workingtree(self):
        # don't distinguish between the branch having a working tree or not
        # when the branch itself is remote.
        self.make_branch_and_tree('.')

        # rely on SFTPServer get_url() pointing at '.'
        self.run_bzr_error(['Already a branch'], ['init', self.get_url()])

    def test_init_append_revisions_only(self):
        self.run_bzr('init --format=dirstate-tags normal_branch6')
        branch = _mod_branch.Branch.open('normal_branch6')
        self.assertEqual(None, branch.get_append_revisions_only())
        self.run_bzr(
            'init --append-revisions-only --format=dirstate-tags branch6')
        branch = _mod_branch.Branch.open('branch6')
        self.assertEqual(True, branch.get_append_revisions_only())
        self.run_bzr_error(['cannot be set to append-revisions-only'],
                           'init --append-revisions-only --format=knit knit')

    def test_init_without_username(self):
        """Ensure init works if username is not set.
        """
        # brz makes user specified whoami mandatory for operations
        # like commit as whoami is recorded. init however is not so final
        # and uses whoami only in a lock file. Without whoami the login name
        # is used. This test is to ensure that init passes even when whoami
        # is not available.
        self.overrideEnv('EMAIL', None)
        self.overrideEnv('BRZ_EMAIL', None)
        out, err = self.run_bzr(['init', 'foo'])
        self.assertEqual(err, '')
        self.assertTrue(os.path.exists('foo'))



# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import re
import sys
from typing import Type

from ..lazy_import import lazy_import

lazy_import(globals(), """
import contextlib
import time

from breezy import (
    config,
    debug,
    graph,
    osutils,
    transactions,
    ui,
    )
from breezy.bzr import (
    pack,
    )
from breezy.bzr.index import (
    CombinedGraphIndex,
    )
""")
from .. import errors, lockable_files, lockdir
from .. import transport as _mod_transport
from ..bzr import btree_index, index
from ..decorators import only_raises
from ..lock import LogicalLockResult
from ..repository import RepositoryWriteLockResult, _LazyListJoin
from ..trace import mutter, note, warning
from .repository import MetaDirRepository, RepositoryFormatMetaDir
from .serializer import Serializer
from .vf_repository import (MetaDirVersionedFileRepository,
                            MetaDirVersionedFileRepositoryFormat,
                            VersionedFileCommitBuilder)


class RetryWithNewPacks(errors.BzrError):
    """Raised when we realize that the packs on disk have changed.

    This is meant as more of a signaling exception, to trap between where a
    local error occurred and the code that can actually handle the error and
    code that can retry appropriately.
    """

    internal_error = True

    _fmt = ("Pack files have changed, reload and retry. context: %(context)s"
            " %(orig_error)s")

    def __init__(self, context, reload_occurred, exc_info):
        """create a new RetryWithNewPacks error.

        :param reload_occurred: Set to True if we know that the packs have
            already been reloaded, and we are failing because of an in-memory
            cache miss. If set to True then we will ignore if a reload says
            nothing has changed, because we assume it has already reloaded. If
            False, then a reload with nothing changed will force an error.
        :param exc_info: The original exception traceback, so if there is a
            problem we can raise the original error (value from sys.exc_info())
        """
        errors.BzrError.__init__(self)
        self.context = context
        self.reload_occurred = reload_occurred
        self.exc_info = exc_info
        self.orig_error = exc_info[1]
        # TODO: The global error handler should probably treat this by
        #       raising/printing the original exception with a bit about
        #       RetryWithNewPacks also not being caught


class RetryAutopack(RetryWithNewPacks):
    """Raised when we are autopacking and we find a missing file.

    Meant as a signaling exception, to tell the autopack code it should try
    again.
    """

    internal_error = True

    _fmt = ("Pack files have changed, reload and try autopack again."
            " context: %(context)s %(orig_error)s")


class PackCommitBuilder(VersionedFileCommitBuilder):
    """Subclass of VersionedFileCommitBuilder to add texts with pack semantics.

    Specifically this uses one knit object rather than one knit object per
    added text, reducing memory and object pressure.
    """

    def __init__(self, repository, parents, config, timestamp=None,
                 timezone=None, committer=None, revprops=None,
                 revision_id=None, lossy=False, owns_transaction=True):
        VersionedFileCommitBuilder.__init__(self, repository, parents, config,
                                            timestamp=timestamp, timezone=timezone, committer=committer,
                                            revprops=revprops, revision_id=revision_id, lossy=lossy,
                                            owns_transaction=owns_transaction)
        self._file_graph = graph.Graph(
            repository._pack_collection.text_index.combined_index)

    def _heads(self, file_id, revision_ids):
        keys = [(file_id, revision_id) for revision_id in revision_ids]
        return {key[1] for key in self._file_graph.heads(keys)}


class Pack:
    """An in memory proxy for a pack and its indices.

    This is a base class that is not directly used, instead the classes
    ExistingPack and NewPack are used.
    """

    # A map of index 'type' to the file extension and position in the
    # index_sizes array.
    index_definitions = {
        'chk': ('.cix', 4),
        'revision': ('.rix', 0),
        'inventory': ('.iix', 1),
        'text': ('.tix', 2),
        'signature': ('.six', 3),
        }

    def __init__(self, revision_index, inventory_index, text_index,
                 signature_index, chk_index=None):
        """Create a pack instance.

        :param revision_index: A GraphIndex for determining what revisions are
            present in the Pack and accessing the locations of their texts.
        :param inventory_index: A GraphIndex for determining what inventories are
            present in the Pack and accessing the locations of their
            texts/deltas.
        :param text_index: A GraphIndex for determining what file texts
            are present in the pack and accessing the locations of their
            texts/deltas (via (fileid, revisionid) tuples).
        :param signature_index: A GraphIndex for determining what signatures are
            present in the Pack and accessing the locations of their texts.
        :param chk_index: A GraphIndex for accessing content by CHK, if the
            pack has one.
        """
        self.revision_index = revision_index
        self.inventory_index = inventory_index
        self.text_index = text_index
        self.signature_index = signature_index
        self.chk_index = chk_index

    def access_tuple(self):
        """Return a tuple (transport, name) for the pack content."""
        return self.pack_transport, self.file_name()

    def _check_references(self):
        """Make sure our external references are present.

        Packs are allowed to have deltas whose base is not in the pack, but it
        must be present somewhere in this collection.  It is not allowed to
        have deltas based on a fallback repository.
        (See <https://bugs.launchpad.net/bzr/+bug/288751>)
        """
        missing_items = {}
        for (index_name, external_refs, index) in [
                ('texts',
                    self._get_external_refs(self.text_index),
                    self._pack_collection.text_index.combined_index),
                ('inventories',
                    self._get_external_refs(self.inventory_index),
                    self._pack_collection.inventory_index.combined_index),
                ]:
            missing = external_refs.difference(
                k for (idx, k, v, r) in
                index.iter_entries(external_refs))
            if missing:
                missing_items[index_name] = sorted(list(missing))
        if missing_items:
            from pprint import pformat
            raise errors.BzrCheckError(
                "Newly created pack file %r has delta references to "
                "items not in its repository:\n%s"
                % (self, pformat(missing_items)))

    def file_name(self):
        """Get the file name for the pack on disk."""
        return self.name + '.pack'

    def get_revision_count(self):
        return self.revision_index.key_count()

    def index_name(self, index_type, name):
        """Get the disk name of an index type for pack name 'name'."""
        return name + Pack.index_definitions[index_type][0]

    def index_offset(self, index_type):
        """Get the position in a index_size array for a given index type."""
        return Pack.index_definitions[index_type][1]

    def inventory_index_name(self, name):
        """The inv index is the name + .iix."""
        return self.index_name('inventory', name)

    def revision_index_name(self, name):
        """The revision index is the name + .rix."""
        return self.index_name('revision', name)

    def signature_index_name(self, name):
        """The signature index is the name + .six."""
        return self.index_name('signature', name)

    def text_index_name(self, name):
        """The text index is the name + .tix."""
        return self.index_name('text', name)

    def _replace_index_with_readonly(self, index_type):
        unlimited_cache = False
        if index_type == 'chk':
            unlimited_cache = True
        index = self.index_class(self.index_transport,
                                 self.index_name(index_type, self.name),
                                 self.index_sizes[self.index_offset(
                                     index_type)],
                                 unlimited_cache=unlimited_cache)
        if index_type == 'chk':
            index._leaf_factory = btree_index._gcchk_factory
        setattr(self, index_type + '_index', index)

    def __lt__(self, other):
        if not isinstance(other, Pack):
            raise TypeError(other)
        return (id(self) < id(other))

    def __hash__(self):
        return hash((type(self), self.revision_index, self.inventory_index,
                     self.text_index, self.signature_index, self.chk_index))


class ExistingPack(Pack):
    """An in memory proxy for an existing .pack and its disk indices."""

    def __init__(self, pack_transport, name, revision_index, inventory_index,
                 text_index, signature_index, chk_index=None):
        """Create an ExistingPack object.

        :param pack_transport: The transport where the pack file resides.
        :param name: The name of the pack on disk in the pack_transport.
        """
        Pack.__init__(self, revision_index, inventory_index, text_index,
                      signature_index, chk_index)
        self.name = name
        self.pack_transport = pack_transport
        if None in (revision_index, inventory_index, text_index,
                    signature_index, name, pack_transport):
            raise AssertionError()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "<{}.{} object at 0x{:x}, {}, {}".format(
            self.__class__.__module__, self.__class__.__name__, id(self),
            self.pack_transport, self.name)

    def __hash__(self):
        return hash((type(self), self.name))


class ResumedPack(ExistingPack):

    def __init__(self, name, revision_index, inventory_index, text_index,
                 signature_index, upload_transport, pack_transport, index_transport,
                 pack_collection, chk_index=None):
        """Create a ResumedPack object."""
        ExistingPack.__init__(self, pack_transport, name, revision_index,
                              inventory_index, text_index, signature_index,
                              chk_index=chk_index)
        self.upload_transport = upload_transport
        self.index_transport = index_transport
        self.index_sizes = [None, None, None, None]
        indices = [
            ('revision', revision_index),
            ('inventory', inventory_index),
            ('text', text_index),
            ('signature', signature_index),
            ]
        if chk_index is not None:
            indices.append(('chk', chk_index))
            self.index_sizes.append(None)
        for index_type, index in indices:
            offset = self.index_offset(index_type)
            self.index_sizes[offset] = index._size
        self.index_class = pack_collection._index_class
        self._pack_collection = pack_collection
        self._state = 'resumed'
        # XXX: perhaps check that the .pack file exists?

    def access_tuple(self):
        if self._state == 'finished':
            return Pack.access_tuple(self)
        elif self._state == 'resumed':
            return self.upload_transport, self.file_name()
        else:
            raise AssertionError(self._state)

    def abort(self):
        self.upload_transport.delete(self.file_name())
        indices = [self.revision_index, self.inventory_index, self.text_index,
                   self.signature_index]
        if self.chk_index is not None:
            indices.append(self.chk_index)
        for index in indices:
            index._transport.delete(index._name)

    def finish(self):
        self._check_references()
        index_types = ['revision', 'inventory', 'text', 'signature']
        if self.chk_index is not None:
            index_types.append('chk')
        for index_type in index_types:
            old_name = self.index_name(index_type, self.name)
            new_name = '../indices/' + old_name
            self.upload_transport.move(old_name, new_name)
            self._replace_index_with_readonly(index_type)
        new_name = '../packs/' + self.file_name()
        self.upload_transport.move(self.file_name(), new_name)
        self._state = 'finished'

    def _get_external_refs(self, index):
        """Return compression parents for this index that are not present.

        This returns any compression parents that are referenced by this index,
        which are not contained *in* this index. They may be present elsewhere.
        """
        return index.external_references(1)


class NewPack(Pack):
    """An in memory proxy for a pack which is being created."""

    def __init__(self, pack_collection, upload_suffix='', file_mode=None):
        """Create a NewPack instance.

        :param pack_collection: A PackCollection into which this is being inserted.
        :param upload_suffix: An optional suffix to be given to any temporary
            files created during the pack creation. e.g '.autopack'
        :param file_mode: Unix permissions for newly created file.
        """
        # The relative locations of the packs are constrained, but all are
        # passed in because the caller has them, so as to avoid object churn.
        index_builder_class = pack_collection._index_builder_class
        if pack_collection.chk_index is not None:
            chk_index = index_builder_class(reference_lists=0)
        else:
            chk_index = None
        Pack.__init__(self,
                      # Revisions: parents list, no text compression.
                      index_builder_class(reference_lists=1),
                      # Inventory: We want to map compression only, but currently the
                      # knit code hasn't been updated enough to understand that, so we
                      # have a regular 2-list index giving parents and compression
                      # source.
                      index_builder_class(reference_lists=2),
                      # Texts: compression and per file graph, for all fileids - so two
                      # reference lists and two elements in the key tuple.
                      index_builder_class(reference_lists=2, key_elements=2),
                      # Signatures: Just blobs to store, no compression, no parents
                      # listing.
                      index_builder_class(reference_lists=0),
                      # CHK based storage - just blobs, no compression or parents.
                      chk_index=chk_index
                      )
        self._pack_collection = pack_collection
        # When we make readonly indices, we need this.
        self.index_class = pack_collection._index_class
        # where should the new pack be opened
        self.upload_transport = pack_collection._upload_transport
        # where are indices written out to
        self.index_transport = pack_collection._index_transport
        # where is the pack renamed to when it is finished?
        self.pack_transport = pack_collection._pack_transport
        # What file mode to upload the pack and indices with.
        self._file_mode = file_mode
        # tracks the content written to the .pack file.
        self._hash = osutils.md5()
        # a tuple with the length in bytes of the indices, once the pack
        # is finalised. (rev, inv, text, sigs, chk_if_in_use)
        self.index_sizes = None
        # How much data to cache when writing packs. Note that this is not
        # synchronised with reads, because it's not in the transport layer, so
        # is not safe unless the client knows it won't be reading from the pack
        # under creation.
        self._cache_limit = 0
        # the temporary pack file name.
        self.random_name = osutils.rand_chars(20) + upload_suffix
        # when was this pack started ?
        self.start_time = time.time()
        # open an output stream for the data added to the pack.
        self.write_stream = self.upload_transport.open_write_stream(
            self.random_name, mode=self._file_mode)
        if 'pack' in debug.debug_flags:
            mutter('%s: create_pack: pack stream open: %s%s t+%6.3fs',
                   time.ctime(), self.upload_transport.base, self.random_name,
                   time.time() - self.start_time)
        # A list of byte sequences to be written to the new pack, and the
        # aggregate size of them.  Stored as a list rather than separate
        # variables so that the _write_data closure below can update them.
        self._buffer = [[], 0]
        # create a callable for adding data
        #
        # robertc says- this is a closure rather than a method on the object
        # so that the variables are locals, and faster than accessing object
        # members.

        def _write_data(bytes, flush=False, _buffer=self._buffer,
                        _write=self.write_stream.write, _update=self._hash.update):
            _buffer[0].append(bytes)
            _buffer[1] += len(bytes)
            # buffer cap
            if _buffer[1] > self._cache_limit or flush:
                bytes = b''.join(_buffer[0])
                _write(bytes)
                _update(bytes)
                _buffer[:] = [[], 0]
        # expose this on self, for the occasion when clients want to add data.
        self._write_data = _write_data
        # a pack writer object to serialise pack records.
        self._writer = pack.ContainerWriter(self._write_data)
        self._writer.begin()
        # what state is the pack in? (open, finished, aborted)
        self._state = 'open'
        # no name until we finish writing the content
        self.name = None

    def abort(self):
        """Cancel creating this pack."""
        self._state = 'aborted'
        self.write_stream.close()
        # Remove the temporary pack file.
        self.upload_transport.delete(self.random_name)
        # The indices have no state on disk.

    def access_tuple(self):
        """Return a tuple (transport, name) for the pack content."""
        if self._state == 'finished':
            return Pack.access_tuple(self)
        elif self._state == 'open':
            return self.upload_transport, self.random_name
        else:
            raise AssertionError(self._state)

    def data_inserted(self):
        """True if data has been added to this pack."""
        return bool(self.get_revision_count() or
                    self.inventory_index.key_count() or
                    self.text_index.key_count() or
                    self.signature_index.key_count() or
                    (self.chk_index is not None and self.chk_index.key_count()))

    def finish_content(self):
        if self.name is not None:
            return
        self._writer.end()
        if self._buffer[1]:
            self._write_data(b'', flush=True)
        self.name = self._hash.hexdigest()

    def finish(self, suspend=False):
        """Finish the new pack.

        This:
         - finalises the content
         - assigns a name (the md5 of the content, currently)
         - writes out the associated indices
         - renames the pack into place.
         - stores the index size tuple for the pack in the index_sizes
           attribute.
        """
        self.finish_content()
        if not suspend:
            self._check_references()
        # write indices
        # XXX: It'd be better to write them all to temporary names, then
        # rename them all into place, so that the window when only some are
        # visible is smaller.  On the other hand none will be seen until
        # they're in the names list.
        self.index_sizes = [None, None, None, None]
        self._write_index('revision', self.revision_index, 'revision',
                          suspend)
        self._write_index('inventory', self.inventory_index, 'inventory',
                          suspend)
        self._write_index('text', self.text_index, 'file texts', suspend)
        self._write_index('signature', self.signature_index,
                          'revision signatures', suspend)
        if self.chk_index is not None:
            self.index_sizes.append(None)
            self._write_index('chk', self.chk_index,
                              'content hash bytes', suspend)
        self.write_stream.close(
            want_fdatasync=self._pack_collection.config_stack.get('repository.fdatasync'))
        # Note that this will clobber an existing pack with the same name,
        # without checking for hash collisions. While this is undesirable this
        # is something that can be rectified in a subsequent release. One way
        # to rectify it may be to leave the pack at the original name, writing
        # its pack-names entry as something like 'HASH: index-sizes
        # temporary-name'. Allocate that and check for collisions, if it is
        # collision free then rename it into place. If clients know this scheme
        # they can handle missing-file errors by:
        #  - try for HASH.pack
        #  - try for temporary-name
        #  - refresh the pack-list to see if the pack is now absent
        new_name = self.name + '.pack'
        if not suspend:
            new_name = '../packs/' + new_name
        self.upload_transport.move(self.random_name, new_name)
        self._state = 'finished'
        if 'pack' in debug.debug_flags:
            # XXX: size might be interesting?
            mutter('%s: create_pack: pack finished: %s%s->%s t+%6.3fs',
                   time.ctime(), self.upload_transport.base, self.random_name,
                   new_name, time.time() - self.start_time)

    def flush(self):
        """Flush any current data."""
        if self._buffer[1]:
            bytes = b''.join(self._buffer[0])
            self.write_stream.write(bytes)
            self._hash.update(bytes)
            self._buffer[:] = [[], 0]

    def _get_external_refs(self, index):
        return index._external_references()

    def set_write_cache_size(self, size):
        self._cache_limit = size

    def _write_index(self, index_type, index, label, suspend=False):
        """Write out an index.

        :param index_type: The type of index to write - e.g. 'revision'.
        :param index: The index object to serialise.
        :param label: What label to give the index e.g. 'revision'.
        """
        index_name = self.index_name(index_type, self.name)
        if suspend:
            transport = self.upload_transport
        else:
            transport = self.index_transport
        index_tempfile = index.finish()
        index_bytes = index_tempfile.read()
        write_stream = transport.open_write_stream(index_name,
                                                   mode=self._file_mode)
        write_stream.write(index_bytes)
        write_stream.close(
            want_fdatasync=self._pack_collection.config_stack.get('repository.fdatasync'))
        self.index_sizes[self.index_offset(index_type)] = len(index_bytes)
        if 'pack' in debug.debug_flags:
            # XXX: size might be interesting?
            mutter('%s: create_pack: wrote %s index: %s%s t+%6.3fs',
                   time.ctime(), label, self.upload_transport.base,
                   self.random_name, time.time() - self.start_time)
        # Replace the writable index on this object with a readonly,
        # presently unloaded index. We should alter
        # the index layer to make its finish() error if add_node is
        # subsequently used. RBC
        self._replace_index_with_readonly(index_type)


class AggregateIndex:
    """An aggregated index for the RepositoryPackCollection.

    AggregateIndex is reponsible for managing the PackAccess object,
    Index-To-Pack mapping, and all indices list for a specific type of index
    such as 'revision index'.

    A CombinedIndex provides an index on a single key space built up
    from several on-disk indices.  The AggregateIndex builds on this
    to provide a knit access layer, and allows having up to one writable
    index within the collection.
    """
    # XXX: Probably 'can be written to' could/should be separated from 'acts
    # like a knit index' -- mbp 20071024

    def __init__(self, reload_func=None, flush_func=None):
        """Create an AggregateIndex.

        :param reload_func: A function to call if we find we are missing an
            index. Should have the form reload_func() => True if the list of
            active pack files has changed.
        """
        self._reload_func = reload_func
        self.index_to_pack = {}
        self.combined_index = CombinedGraphIndex([], reload_func=reload_func)
        self.data_access = _DirectPackAccess(self.index_to_pack,
                                             reload_func=reload_func,
                                             flush_func=flush_func)
        self.add_callback = None

    def add_index(self, index, pack):
        """Add index to the aggregate, which is an index for Pack pack.

        Future searches on the aggregate index will seach this new index
        before all previously inserted indices.

        :param index: An Index for the pack.
        :param pack: A Pack instance.
        """
        # expose it to the index map
        self.index_to_pack[index] = pack.access_tuple()
        # put it at the front of the linear index list
        self.combined_index.insert_index(0, index, pack.name)

    def add_writable_index(self, index, pack):
        """Add an index which is able to have data added to it.

        There can be at most one writable index at any time.  Any
        modifications made to the knit are put into this index.

        :param index: An index from the pack parameter.
        :param pack: A Pack instance.
        """
        if self.add_callback is not None:
            raise AssertionError(
                "%s already has a writable index through %s" %
                (self, self.add_callback))
        # allow writing: queue writes to a new index
        self.add_index(index, pack)
        # Updates the index to packs mapping as a side effect,
        self.data_access.set_writer(pack._writer, index, pack.access_tuple())
        self.add_callback = index.add_nodes

    def clear(self):
        """Reset all the aggregate data to nothing."""
        self.data_access.set_writer(None, None, (None, None))
        self.index_to_pack.clear()
        del self.combined_index._indices[:]
        del self.combined_index._index_names[:]
        self.add_callback = None

    def remove_index(self, index):
        """Remove index from the indices used to answer queries.

        :param index: An index from the pack parameter.
        """
        del self.index_to_pack[index]
        pos = self.combined_index._indices.index(index)
        del self.combined_index._indices[pos]
        del self.combined_index._index_names[pos]
        if (self.add_callback is not None and
                getattr(index, 'add_nodes', None) == self.add_callback):
            self.add_callback = None
            self.data_access.set_writer(None, None, (None, None))


class Packer:
    """Create a pack from packs."""

    def __init__(self, pack_collection, packs, suffix, revision_ids=None,
                 reload_func=None):
        """Create a Packer.

        :param pack_collection: A RepositoryPackCollection object where the
            new pack is being written to.
        :param packs: The packs to combine.
        :param suffix: The suffix to use on the temporary files for the pack.
        :param revision_ids: Revision ids to limit the pack to.
        :param reload_func: A function to call if a pack file/index goes
            missing. The side effect of calling this function should be to
            update self.packs. See also AggregateIndex
        """
        self.packs = packs
        self.suffix = suffix
        self.revision_ids = revision_ids
        # The pack object we are creating.
        self.new_pack = None
        self._pack_collection = pack_collection
        self._reload_func = reload_func
        # The index layer keys for the revisions being copied. None for 'all
        # objects'.
        self._revision_keys = None
        # What text keys to copy. None for 'all texts'. This is set by
        # _copy_inventory_texts
        self._text_filter = None

    def pack(self, pb=None):
        """Create a new pack by reading data from other packs.

        This does little more than a bulk copy of data. One key difference
        is that data with the same item key across multiple packs is elided
        from the output. The new pack is written into the current pack store
        along with its indices, and the name added to the pack names. The
        source packs are not altered and are not required to be in the current
        pack collection.

        :param pb: An optional progress bar to use. A nested bar is created if
            this is None.
        :return: A Pack object, or None if nothing was copied.
        """
        # open a pack - using the same name as the last temporary file
        # - which has already been flushed, so it's safe.
        # XXX: - duplicate code warning with start_write_group; fix before
        #      considering 'done'.
        if self._pack_collection._new_pack is not None:
            raise errors.BzrError('call to %s.pack() while another pack is'
                                  ' being written.'
                                  % (self.__class__.__name__,))
        if self.revision_ids is not None:
            if len(self.revision_ids) == 0:
                # silly fetch request.
                return None
            else:
                self.revision_ids = frozenset(self.revision_ids)
                self.revision_keys = frozenset((revid,) for revid in
                                               self.revision_ids)
        if pb is None:
            self.pb = ui.ui_factory.nested_progress_bar()
        else:
            self.pb = pb
        try:
            return self._create_pack_from_packs()
        finally:
            if pb is None:
                self.pb.finished()

    def open_pack(self):
        """Open a pack for the pack we are creating."""
        new_pack = self._pack_collection.pack_factory(self._pack_collection,
                                                      upload_suffix=self.suffix,
                                                      file_mode=self._pack_collection.repo.controldir._get_file_mode())
        # We know that we will process all nodes in order, and don't need to
        # query, so don't combine any indices spilled to disk until we are done
        new_pack.revision_index.set_optimize(combine_backing_indices=False)
        new_pack.inventory_index.set_optimize(combine_backing_indices=False)
        new_pack.text_index.set_optimize(combine_backing_indices=False)
        new_pack.signature_index.set_optimize(combine_backing_indices=False)
        return new_pack

    def _copy_revision_texts(self):
        """Copy revision data to the new pack."""
        raise NotImplementedError(self._copy_revision_texts)

    def _copy_inventory_texts(self):
        """Copy the inventory texts to the new pack.

        self._revision_keys is used to determine what inventories to copy.

        Sets self._text_filter appropriately.
        """
        raise NotImplementedError(self._copy_inventory_texts)

    def _copy_text_texts(self):
        raise NotImplementedError(self._copy_text_texts)

    def _create_pack_from_packs(self):
        raise NotImplementedError(self._create_pack_from_packs)

    def _log_copied_texts(self):
        if 'pack' in debug.debug_flags:
            mutter('%s: create_pack: file texts copied: %s%s %d items t+%6.3fs',
                   time.ctime(), self._pack_collection._upload_transport.base,
                   self.new_pack.random_name,
                   self.new_pack.text_index.key_count(),
                   time.time() - self.new_pack.start_time)

    def _use_pack(self, new_pack):
        """Return True if new_pack should be used.

        :param new_pack: The pack that has just been created.
        :return: True if the pack should be used.
        """
        return new_pack.data_inserted()


class RepositoryPackCollection:
    """Management of packs within a repository.

    :ivar _names: map of {pack_name: (index_size,)}
    """

    pack_factory: Type[NewPack]
    resumed_pack_factory: Type[ResumedPack]
    normal_packer_class: Type[Packer]
    optimising_packer_class: Type[Packer]

    def __init__(self, repo, transport, index_transport, upload_transport,
                 pack_transport, index_builder_class, index_class,
                 use_chk_index):
        """Create a new RepositoryPackCollection.

        :param transport: Addresses the repository base directory
            (typically .bzr/repository/).
        :param index_transport: Addresses the directory containing indices.
        :param upload_transport: Addresses the directory into which packs are written
            while they're being created.
        :param pack_transport: Addresses the directory of existing complete packs.
        :param index_builder_class: The index builder class to use.
        :param index_class: The index class to use.
        :param use_chk_index: Whether to setup and manage a CHK index.
        """
        # XXX: This should call self.reset()
        self.repo = repo
        self.transport = transport
        self._index_transport = index_transport
        self._upload_transport = upload_transport
        self._pack_transport = pack_transport
        self._index_builder_class = index_builder_class
        self._index_class = index_class
        self._suffix_offsets = {'.rix': 0, '.iix': 1, '.tix': 2, '.six': 3,
                                '.cix': 4}
        self.packs = []
        # name:Pack mapping
        self._names = None
        self._packs_by_name = {}
        # the previous pack-names content
        self._packs_at_load = None
        # when a pack is being created by this object, the state of that pack.
        self._new_pack = None
        # aggregated revision index data
        flush = self._flush_new_pack
        self.revision_index = AggregateIndex(self.reload_pack_names, flush)
        self.inventory_index = AggregateIndex(self.reload_pack_names, flush)
        self.text_index = AggregateIndex(self.reload_pack_names, flush)
        self.signature_index = AggregateIndex(self.reload_pack_names, flush)
        all_indices = [self.revision_index, self.inventory_index,
                       self.text_index, self.signature_index]
        if use_chk_index:
            self.chk_index = AggregateIndex(self.reload_pack_names, flush)
            all_indices.append(self.chk_index)
        else:
            # used to determine if we're using a chk_index elsewhere.
            self.chk_index = None
        # Tell all the CombinedGraphIndex objects about each other, so they can
        # share hints about which pack names to search first.
        all_combined = [agg_idx.combined_index for agg_idx in all_indices]
        for combined_idx in all_combined:
            combined_idx.set_sibling_indices(
                set(all_combined).difference([combined_idx]))
        # resumed packs
        self._resumed_packs = []
        self.config_stack = config.LocationStack(self.transport.base)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.repo)

    def add_pack_to_memory(self, pack):
        """Make a Pack object available to the repository to satisfy queries.

        :param pack: A Pack object.
        """
        if pack.name in self._packs_by_name:
            raise AssertionError(
                'pack {} already in _packs_by_name'.format(pack.name))
        self.packs.append(pack)
        self._packs_by_name[pack.name] = pack
        self.revision_index.add_index(pack.revision_index, pack)
        self.inventory_index.add_index(pack.inventory_index, pack)
        self.text_index.add_index(pack.text_index, pack)
        self.signature_index.add_index(pack.signature_index, pack)
        if self.chk_index is not None:
            self.chk_index.add_index(pack.chk_index, pack)

    def all_packs(self):
        """Return a list of all the Pack objects this repository has.

        Note that an in-progress pack being created is not returned.

        :return: A list of Pack objects for all the packs in the repository.
        """
        result = []
        for name in self.names():
            result.append(self.get_pack_by_name(name))
        return result

    def autopack(self):
        """Pack the pack collection incrementally.

        This will not attempt global reorganisation or recompression,
        rather it will just ensure that the total number of packs does
        not grow without bound. It uses the _max_pack_count method to
        determine if autopacking is needed, and the pack_distribution
        method to determine the number of revisions in each pack.

        If autopacking takes place then the packs name collection will have
        been flushed to disk - packing requires updating the name collection
        in synchronisation with certain steps. Otherwise the names collection
        is not flushed.

        :return: Something evaluating true if packing took place.
        """
        while True:
            try:
                return self._do_autopack()
            except RetryAutopack:
                # If we get a RetryAutopack exception, we should abort the
                # current action, and retry.
                pass

    def _do_autopack(self):
        # XXX: Should not be needed when the management of indices is sane.
        total_revisions = self.revision_index.combined_index.key_count()
        total_packs = len(self._names)
        if self._max_pack_count(total_revisions) >= total_packs:
            return None
        # determine which packs need changing
        pack_distribution = self.pack_distribution(total_revisions)
        existing_packs = []
        for pack in self.all_packs():
            revision_count = pack.get_revision_count()
            if revision_count == 0:
                # revision less packs are not generated by normal operation,
                # only by operations like sign-my-commits, and thus will not
                # tend to grow rapdily or without bound like commit containing
                # packs do - leave them alone as packing them really should
                # group their data with the relevant commit, and that may
                # involve rewriting ancient history - which autopack tries to
                # avoid. Alternatively we could not group the data but treat
                # each of these as having a single revision, and thus add
                # one revision for each to the total revision count, to get
                # a matching distribution.
                continue
            existing_packs.append((revision_count, pack))
        pack_operations = self.plan_autopack_combinations(
            existing_packs, pack_distribution)
        num_new_packs = len(pack_operations)
        num_old_packs = sum([len(po[1]) for po in pack_operations])
        num_revs_affected = sum([po[0] for po in pack_operations])
        mutter('Auto-packing repository %s, which has %d pack files, '
               'containing %d revisions. Packing %d files into %d affecting %d'
               ' revisions', str(
                   self), total_packs, total_revisions, num_old_packs,
               num_new_packs, num_revs_affected)
        result = self._execute_pack_operations(pack_operations, packer_class=self.normal_packer_class,
                                               reload_func=self._restart_autopack)
        mutter('Auto-packing repository %s completed', str(self))
        return result

    def _execute_pack_operations(self, pack_operations, packer_class,
                                 reload_func=None):
        """Execute a series of pack operations.

        :param pack_operations: A list of [revision_count, packs_to_combine].
        :param packer_class: The class of packer to use
        :return: The new pack names.
        """
        for revision_count, packs in pack_operations:
            # we may have no-ops from the setup logic
            if len(packs) == 0:
                continue
            packer = packer_class(self, packs, '.autopack',
                                  reload_func=reload_func)
            try:
                result = packer.pack()
            except RetryWithNewPacks:
                # An exception is propagating out of this context, make sure
                # this packer has cleaned up. Packer() doesn't set its new_pack
                # state into the RepositoryPackCollection object, so we only
                # have access to it directly here.
                if packer.new_pack is not None:
                    packer.new_pack.abort()
                raise
            if result is None:
                return
            for pack in packs:
                self._remove_pack_from_memory(pack)
        # record the newly available packs and stop advertising the old
        # packs
        to_be_obsoleted = []
        for _, packs in pack_operations:
            to_be_obsoleted.extend(packs)
        result = self._save_pack_names(clear_obsolete_packs=True,
                                       obsolete_packs=to_be_obsoleted)
        return result

    def _flush_new_pack(self):
        if self._new_pack is not None:
            self._new_pack.flush()

    def lock_names(self):
        """Acquire the mutex around the pack-names index.

        This cannot be used in the middle of a read-only transaction on the
        repository.
        """
        self.repo.control_files.lock_write()

    def _already_packed(self):
        """Is the collection already packed?"""
        return not (self.repo._format.pack_compresses or (len(self._names) > 1))

    def pack(self, hint=None, clean_obsolete_packs=False):
        """Pack the pack collection totally."""
        self.ensure_loaded()
        total_packs = len(self._names)
        if self._already_packed():
            return
        total_revisions = self.revision_index.combined_index.key_count()
        # XXX: the following may want to be a class, to pack with a given
        # policy.
        mutter('Packing repository %s, which has %d pack files, '
               'containing %d revisions with hint %r.', str(self), total_packs,
               total_revisions, hint)
        while True:
            try:
                self._try_pack_operations(hint)
            except RetryPackOperations:
                continue
            break

        if clean_obsolete_packs:
            self._clear_obsolete_packs()

    def _try_pack_operations(self, hint):
        """Calculate the pack operations based on the hint (if any), and
        execute them.
        """
        # determine which packs need changing
        pack_operations = [[0, []]]
        for pack in self.all_packs():
            if hint is None or pack.name in hint:
                # Either no hint was provided (so we are packing everything),
                # or this pack was included in the hint.
                pack_operations[-1][0] += pack.get_revision_count()
                pack_operations[-1][1].append(pack)
        self._execute_pack_operations(pack_operations,
                                      packer_class=self.optimising_packer_class,
                                      reload_func=self._restart_pack_operations)

    def plan_autopack_combinations(self, existing_packs, pack_distribution):
        """Plan a pack operation.

        :param existing_packs: The packs to pack. (A list of (revcount, Pack)
            tuples).
        :param pack_distribution: A list with the number of revisions desired
            in each pack.
        """
        if len(existing_packs) <= len(pack_distribution):
            return []
        existing_packs.sort(reverse=True)
        pack_operations = [[0, []]]
        # plan out what packs to keep, and what to reorganise
        while len(existing_packs):
            # take the largest pack, and if it's less than the head of the
            # distribution chart we will include its contents in the new pack
            # for that position. If it's larger, we remove its size from the
            # distribution chart
            next_pack_rev_count, next_pack = existing_packs.pop(0)
            if next_pack_rev_count >= pack_distribution[0]:
                # this is already packed 'better' than this, so we can
                # not waste time packing it.
                while next_pack_rev_count > 0:
                    next_pack_rev_count -= pack_distribution[0]
                    if next_pack_rev_count >= 0:
                        # more to go
                        del pack_distribution[0]
                    else:
                        # didn't use that entire bucket up
                        pack_distribution[0] = -next_pack_rev_count
            else:
                # add the revisions we're going to add to the next output pack
                pack_operations[-1][0] += next_pack_rev_count
                # allocate this pack to the next pack sub operation
                pack_operations[-1][1].append(next_pack)
                if pack_operations[-1][0] >= pack_distribution[0]:
                    # this pack is used up, shift left.
                    del pack_distribution[0]
                    pack_operations.append([0, []])
        # Now that we know which pack files we want to move, shove them all
        # into a single pack file.
        final_rev_count = 0
        final_pack_list = []
        for num_revs, pack_files in pack_operations:
            final_rev_count += num_revs
            final_pack_list.extend(pack_files)
        if len(final_pack_list) == 1:
            raise AssertionError('We somehow generated an autopack with a'
                                 ' single pack file being moved.')
            return []
        return [[final_rev_count, final_pack_list]]

    def ensure_loaded(self):
        """Ensure we have read names from disk.

        :return: True if the disk names had not been previously read.
        """
        # NB: if you see an assertion error here, it's probably access against
        # an unlocked repo. Naughty.
        if not self.repo.is_locked():
            raise errors.ObjectNotLocked(self.repo)
        if self._names is None:
            self._names = {}
            self._packs_at_load = set()
            for index, key, value in self._iter_disk_pack_index():
                name = key[0].decode('ascii')
                self._names[name] = self._parse_index_sizes(value)
                self._packs_at_load.add((name, value))
            result = True
        else:
            result = False
        # populate all the metadata.
        self.all_packs()
        return result

    def _parse_index_sizes(self, value):
        """Parse a string of index sizes."""
        return tuple(int(digits) for digits in value.split(b' '))

    def get_pack_by_name(self, name):
        """Get a Pack object by name.

        :param name: The name of the pack - e.g. '123456'
        :return: A Pack object.
        """
        try:
            return self._packs_by_name[name]
        except KeyError:
            rev_index = self._make_index(name, '.rix')
            inv_index = self._make_index(name, '.iix')
            txt_index = self._make_index(name, '.tix')
            sig_index = self._make_index(name, '.six')
            if self.chk_index is not None:
                chk_index = self._make_index(name, '.cix', is_chk=True)
            else:
                chk_index = None
            result = ExistingPack(self._pack_transport, name, rev_index,
                                  inv_index, txt_index, sig_index, chk_index)
            self.add_pack_to_memory(result)
            return result

    def _resume_pack(self, name):
        """Get a suspended Pack object by name.

        :param name: The name of the pack - e.g. '123456'
        :return: A Pack object.
        """
        if not re.match('[a-f0-9]{32}', name):
            # Tokens should be md5sums of the suspended pack file, i.e. 32 hex
            # digits.
            raise errors.UnresumableWriteGroup(
                self.repo, [name], 'Malformed write group token')
        try:
            rev_index = self._make_index(name, '.rix', resume=True)
            inv_index = self._make_index(name, '.iix', resume=True)
            txt_index = self._make_index(name, '.tix', resume=True)
            sig_index = self._make_index(name, '.six', resume=True)
            if self.chk_index is not None:
                chk_index = self._make_index(name, '.cix', resume=True,
                                             is_chk=True)
            else:
                chk_index = None
            result = self.resumed_pack_factory(name, rev_index, inv_index,
                                               txt_index, sig_index, self._upload_transport,
                                               self._pack_transport, self._index_transport, self,
                                               chk_index=chk_index)
        except _mod_transport.NoSuchFile as e:
            raise errors.UnresumableWriteGroup(self.repo, [name], str(e))
        self.add_pack_to_memory(result)
        self._resumed_packs.append(result)
        return result

    def allocate(self, a_new_pack):
        """Allocate name in the list of packs.

        :param a_new_pack: A NewPack instance to be added to the collection of
            packs for this repository.
        """
        self.ensure_loaded()
        if a_new_pack.name in self._names:
            raise errors.BzrError(
                'Pack {!r} already exists in {}'.format(a_new_pack.name, self))
        self._names[a_new_pack.name] = tuple(a_new_pack.index_sizes)
        self.add_pack_to_memory(a_new_pack)

    def _iter_disk_pack_index(self):
        """Iterate over the contents of the pack-names index.

        This is used when loading the list from disk, and before writing to
        detect updates from others during our write operation.
        :return: An iterator of the index contents.
        """
        return self._index_class(self.transport, 'pack-names', None
                                 ).iter_all_entries()

    def _make_index(self, name, suffix, resume=False, is_chk=False):
        size_offset = self._suffix_offsets[suffix]
        index_name = name + suffix
        if resume:
            transport = self._upload_transport
            index_size = transport.stat(index_name).st_size
        else:
            transport = self._index_transport
            index_size = self._names[name][size_offset]
        index = self._index_class(transport, index_name, index_size,
                                  unlimited_cache=is_chk)
        if is_chk and self._index_class is btree_index.BTreeGraphIndex:
            index._leaf_factory = btree_index._gcchk_factory
        return index

    def _max_pack_count(self, total_revisions):
        """Return the maximum number of packs to use for total revisions.

        :param total_revisions: The total number of revisions in the
            repository.
        """
        if not total_revisions:
            return 1
        digits = str(total_revisions)
        result = 0
        for digit in digits:
            result += int(digit)
        return result

    def names(self):
        """Provide an order to the underlying names."""
        return sorted(self._names.keys())

    def _obsolete_packs(self, packs):
        """Move a number of packs which have been obsoleted out of the way.

        Each pack and its associated indices are moved out of the way.

        Note: for correctness this function should only be called after a new
        pack names index has been written without these pack names, and with
        the names of packs that contain the data previously available via these
        packs.

        :param packs: The packs to obsolete.
        :param return: None.
        """
        for pack in packs:
            try:
                try:
                    pack.pack_transport.move(pack.file_name(),
                                             '../obsolete_packs/' + pack.file_name())
                except _mod_transport.NoSuchFile:
                    # perhaps obsolete_packs was removed? Let's create it and
                    # try again
                    try:
                        pack.pack_transport.mkdir('../obsolete_packs/')
                    except _mod_transport.FileExists:
                        pass
                    pack.pack_transport.move(pack.file_name(),
                                             '../obsolete_packs/' + pack.file_name())
            except (errors.PathError, errors.TransportError) as e:
                # TODO: Should these be warnings or mutters?
                mutter("couldn't rename obsolete pack, skipping it:\n%s"
                       % (e,))
            # TODO: Probably needs to know all possible indices for this pack
            # - or maybe list the directory and move all indices matching this
            # name whether we recognize it or not?
            suffixes = ['.iix', '.six', '.tix', '.rix']
            if self.chk_index is not None:
                suffixes.append('.cix')
            for suffix in suffixes:
                try:
                    self._index_transport.move(pack.name + suffix,
                                               '../obsolete_packs/' + pack.name + suffix)
                except (errors.PathError, errors.TransportError) as e:
                    mutter("couldn't rename obsolete index, skipping it:\n%s"
                           % (e,))

    def pack_distribution(self, total_revisions):
        """Generate a list of the number of revisions to put in each pack.

        :param total_revisions: The total number of revisions in the
            repository.
        """
        if total_revisions == 0:
            return [0]
        digits = reversed(str(total_revisions))
        result = []
        for exponent, count in enumerate(digits):
            size = 10 ** exponent
            for pos in range(int(count)):
                result.append(size)
        return list(reversed(result))

    def _pack_tuple(self, name):
        """Return a tuple with the transport and file name for a pack name."""
        return self._pack_transport, name + '.pack'

    def _remove_pack_from_memory(self, pack):
        """Remove pack from the packs accessed by this repository.

        Only affects memory state, until self._save_pack_names() is invoked.
        """
        self._names.pop(pack.name)
        self._packs_by_name.pop(pack.name)
        self._remove_pack_indices(pack)
        self.packs.remove(pack)

    def _remove_pack_indices(self, pack, ignore_missing=False):
        """Remove the indices for pack from the aggregated indices.

        :param ignore_missing: Suppress KeyErrors from calling remove_index.
        """
        for index_type in Pack.index_definitions:
            attr_name = index_type + '_index'
            aggregate_index = getattr(self, attr_name)
            if aggregate_index is not None:
                pack_index = getattr(pack, attr_name)
                try:
                    aggregate_index.remove_index(pack_index)
                except KeyError:
                    if ignore_missing:
                        continue
                    raise

    def reset(self):
        """Clear all cached data."""
        # cached revision data
        self.revision_index.clear()
        # cached signature data
        self.signature_index.clear()
        # cached file text data
        self.text_index.clear()
        # cached inventory data
        self.inventory_index.clear()
        # cached chk data
        if self.chk_index is not None:
            self.chk_index.clear()
        # remove the open pack
        self._new_pack = None
        # information about packs.
        self._names = None
        self.packs = []
        self._packs_by_name = {}
        self._packs_at_load = None

    def _unlock_names(self):
        """Release the mutex around the pack-names index."""
        self.repo.control_files.unlock()

    def _diff_pack_names(self):
        """Read the pack names from disk, and compare it to the one in memory.

        :return: (disk_nodes, deleted_nodes, new_nodes)
            disk_nodes    The final set of nodes that should be referenced
            deleted_nodes Nodes which have been removed from when we started
            new_nodes     Nodes that are newly introduced
        """
        # load the disk nodes across
        disk_nodes = set()
        for index, key, value in self._iter_disk_pack_index():
            disk_nodes.add((key[0].decode('ascii'), value))
        orig_disk_nodes = set(disk_nodes)

        # do a two-way diff against our original content
        current_nodes = set()
        for name, sizes in self._names.items():
            current_nodes.add(
                (name, b' '.join(b'%d' % size for size in sizes)))

        # Packs no longer present in the repository, which were present when we
        # locked the repository
        deleted_nodes = self._packs_at_load - current_nodes
        # Packs which this process is adding
        new_nodes = current_nodes - self._packs_at_load

        # Update the disk_nodes set to include the ones we are adding, and
        # remove the ones which were removed by someone else
        disk_nodes.difference_update(deleted_nodes)
        disk_nodes.update(new_nodes)

        return disk_nodes, deleted_nodes, new_nodes, orig_disk_nodes

    def _syncronize_pack_names_from_disk_nodes(self, disk_nodes):
        """Given the correct set of pack files, update our saved info.

        :return: (removed, added, modified)
            removed     pack names removed from self._names
            added       pack names added to self._names
            modified    pack names that had changed value
        """
        removed = []
        added = []
        modified = []
        ## self._packs_at_load = disk_nodes
        new_names = dict(disk_nodes)
        # drop no longer present nodes
        for pack in self.all_packs():
            if pack.name not in new_names:
                removed.append(pack.name)
                self._remove_pack_from_memory(pack)
        # add new nodes/refresh existing ones
        for name, value in disk_nodes:
            sizes = self._parse_index_sizes(value)
            if name in self._names:
                # existing
                if sizes != self._names[name]:
                    # the pack for name has had its indices replaced - rare but
                    # important to handle. XXX: probably can never happen today
                    # because the three-way merge code above does not handle it
                    # - you may end up adding the same key twice to the new
                    # disk index because the set values are the same, unless
                    # the only index shows up as deleted by the set difference
                    # - which it may. Until there is a specific test for this,
                    # assume it's broken. RBC 20071017.
                    self._remove_pack_from_memory(self.get_pack_by_name(name))
                    self._names[name] = sizes
                    self.get_pack_by_name(name)
                    modified.append(name)
            else:
                # new
                self._names[name] = sizes
                self.get_pack_by_name(name)
                added.append(name)
        return removed, added, modified

    def _save_pack_names(self, clear_obsolete_packs=False, obsolete_packs=None):
        """Save the list of packs.

        This will take out the mutex around the pack names list for the
        duration of the method call. If concurrent updates have been made, a
        three-way merge between the current list and the current in memory list
        is performed.

        :param clear_obsolete_packs: If True, clear out the contents of the
            obsolete_packs directory.
        :param obsolete_packs: Packs that are obsolete once the new pack-names
            file has been written.
        :return: A list of the names saved that were not previously on disk.
        """
        already_obsolete = []
        self.lock_names()
        try:
            builder = self._index_builder_class()
            (disk_nodes, deleted_nodes, new_nodes,
             orig_disk_nodes) = self._diff_pack_names()
            # TODO: handle same-name, index-size-changes here -
            # e.g. use the value from disk, not ours, *unless* we're the one
            # changing it.
            for name, value in disk_nodes:
                builder.add_node((name.encode('ascii'), ), value)
            self.transport.put_file('pack-names', builder.finish(),
                                    mode=self.repo.controldir._get_file_mode())
            self._packs_at_load = disk_nodes
            if clear_obsolete_packs:
                to_preserve = None
                if obsolete_packs:
                    to_preserve = {o.name for o in obsolete_packs}
                already_obsolete = self._clear_obsolete_packs(to_preserve)
        finally:
            self._unlock_names()
        # synchronise the memory packs list with what we just wrote:
        self._syncronize_pack_names_from_disk_nodes(disk_nodes)
        if obsolete_packs:
            # TODO: We could add one more condition here. "if o.name not in
            #       orig_disk_nodes and o != the new_pack we haven't written to
            #       disk yet. However, the new pack object is not easily
            #       accessible here (it would have to be passed through the
            #       autopacking code, etc.)
            obsolete_packs = [o for o in obsolete_packs
                              if o.name not in already_obsolete]
            self._obsolete_packs(obsolete_packs)
        return [new_node[0] for new_node in new_nodes]

    def reload_pack_names(self):
        """Sync our pack listing with what is present in the repository.

        This should be called when we find out that something we thought was
        present is now missing. This happens when another process re-packs the
        repository, etc.

        :return: True if the in-memory list of packs has been altered at all.
        """
        # The ensure_loaded call is to handle the case where the first call
        # made involving the collection was to reload_pack_names, where we
        # don't have a view of disk contents. It's a bit of a bandaid, and
        # causes two reads of pack-names, but it's a rare corner case not
        # struck with regular push/pull etc.
        first_read = self.ensure_loaded()
        if first_read:
            return True
        # out the new value.
        (disk_nodes, deleted_nodes, new_nodes,
         orig_disk_nodes) = self._diff_pack_names()
        # _packs_at_load is meant to be the explicit list of names in
        # 'pack-names' at then start. As such, it should not contain any
        # pending names that haven't been written out yet.
        self._packs_at_load = orig_disk_nodes
        (removed, added,
         modified) = self._syncronize_pack_names_from_disk_nodes(disk_nodes)
        if removed or added or modified:
            return True
        return False

    def _restart_autopack(self):
        """Reload the pack names list, and restart the autopack code."""
        if not self.reload_pack_names():
            # Re-raise the original exception, because something went missing
            # and a restart didn't find it
            raise
        raise RetryAutopack(self.repo, False, sys.exc_info())

    def _restart_pack_operations(self):
        """Reload the pack names list, and restart the autopack code."""
        if not self.reload_pack_names():
            # Re-raise the original exception, because something went missing
            # and a restart didn't find it
            raise
        raise RetryPackOperations(self.repo, False, sys.exc_info())

    def _clear_obsolete_packs(self, preserve=None):
        """Delete everything from the obsolete-packs directory.

        :return: A list of pack identifiers (the filename without '.pack') that
            were found in obsolete_packs.
        """
        found = []
        obsolete_pack_transport = self.transport.clone('obsolete_packs')
        if preserve is None:
            preserve = set()
        try:
            obsolete_pack_files = obsolete_pack_transport.list_dir('.')
        except _mod_transport.NoSuchFile:
            return found
        for filename in obsolete_pack_files:
            name, ext = osutils.splitext(filename)
            if ext == '.pack':
                found.append(name)
            if name in preserve:
                continue
            try:
                obsolete_pack_transport.delete(filename)
            except (errors.PathError, errors.TransportError) as e:
                warning("couldn't delete obsolete pack, skipping it:\n%s"
                        % (e,))
        return found

    def _start_write_group(self):
        # Do not permit preparation for writing if we're not in a 'write lock'.
        if not self.repo.is_write_locked():
            raise errors.NotWriteLocked(self)
        self._new_pack = self.pack_factory(self, upload_suffix='.pack',
                                           file_mode=self.repo.controldir._get_file_mode())
        # allow writing: queue writes to a new index
        self.revision_index.add_writable_index(self._new_pack.revision_index,
                                               self._new_pack)
        self.inventory_index.add_writable_index(self._new_pack.inventory_index,
                                                self._new_pack)
        self.text_index.add_writable_index(self._new_pack.text_index,
                                           self._new_pack)
        self._new_pack.text_index.set_optimize(combine_backing_indices=False)
        self.signature_index.add_writable_index(self._new_pack.signature_index,
                                                self._new_pack)
        if self.chk_index is not None:
            self.chk_index.add_writable_index(self._new_pack.chk_index,
                                              self._new_pack)
            self.repo.chk_bytes._index._add_callback = self.chk_index.add_callback
            self._new_pack.chk_index.set_optimize(
                combine_backing_indices=False)

        self.repo.inventories._index._add_callback = self.inventory_index.add_callback
        self.repo.revisions._index._add_callback = self.revision_index.add_callback
        self.repo.signatures._index._add_callback = self.signature_index.add_callback
        self.repo.texts._index._add_callback = self.text_index.add_callback

    def _abort_write_group(self):
        # FIXME: just drop the transient index.
        # forget what names there are
        if self._new_pack is not None:
            with contextlib.ExitStack() as stack:
                stack.callback(setattr, self, '_new_pack', None)
                # If we aborted while in the middle of finishing the write
                # group, _remove_pack_indices could fail because the indexes are
                # already gone.  But they're not there we shouldn't fail in this
                # case, so we pass ignore_missing=True.
                stack.callback(self._remove_pack_indices, self._new_pack,
                               ignore_missing=True)
                self._new_pack.abort()
        for resumed_pack in self._resumed_packs:
            with contextlib.ExitStack() as stack:
                # See comment in previous finally block.
                stack.callback(self._remove_pack_indices, resumed_pack,
                               ignore_missing=True)
                resumed_pack.abort()
        del self._resumed_packs[:]

    def _remove_resumed_pack_indices(self):
        for resumed_pack in self._resumed_packs:
            self._remove_pack_indices(resumed_pack)
        del self._resumed_packs[:]

    def _check_new_inventories(self):
        """Detect missing inventories in this write group.

        :returns: list of strs, summarising any problems found.  If the list is
            empty no problems were found.
        """
        # The base implementation does no checks.  GCRepositoryPackCollection
        # overrides this.
        return []

    def _commit_write_group(self):
        all_missing = set()
        for prefix, versioned_file in (
                ('revisions', self.repo.revisions),
                ('inventories', self.repo.inventories),
                ('texts', self.repo.texts),
                ('signatures', self.repo.signatures),
                ):
            missing = versioned_file.get_missing_compression_parent_keys()
            all_missing.update([(prefix,) + key for key in missing])
        if all_missing:
            raise errors.BzrCheckError(
                "Repository %s has missing compression parent(s) %r "
                % (self.repo, sorted(all_missing)))
        problems = self._check_new_inventories()
        if problems:
            problems_summary = '\n'.join(problems)
            raise errors.BzrCheckError(
                "Cannot add revision(s) to repository: " + problems_summary)
        self._remove_pack_indices(self._new_pack)
        any_new_content = False
        if self._new_pack.data_inserted():
            # get all the data to disk and read to use
            self._new_pack.finish()
            self.allocate(self._new_pack)
            self._new_pack = None
            any_new_content = True
        else:
            self._new_pack.abort()
            self._new_pack = None
        for resumed_pack in self._resumed_packs:
            # XXX: this is a pretty ugly way to turn the resumed pack into a
            # properly committed pack.
            self._names[resumed_pack.name] = None
            self._remove_pack_from_memory(resumed_pack)
            resumed_pack.finish()
            self.allocate(resumed_pack)
            any_new_content = True
        del self._resumed_packs[:]
        if any_new_content:
            result = self.autopack()
            if not result:
                # when autopack takes no steps, the names list is still
                # unsaved.
                return self._save_pack_names()
            return result
        return []

    def _suspend_write_group(self):
        tokens = [pack.name for pack in self._resumed_packs]
        self._remove_pack_indices(self._new_pack)
        if self._new_pack.data_inserted():
            # get all the data to disk and read to use
            self._new_pack.finish(suspend=True)
            tokens.append(self._new_pack.name)
            self._new_pack = None
        else:
            self._new_pack.abort()
            self._new_pack = None
        self._remove_resumed_pack_indices()
        return tokens

    def _resume_write_group(self, tokens):
        for token in tokens:
            self._resume_pack(token)


class PackRepository(MetaDirVersionedFileRepository):
    """Repository with knit objects stored inside pack containers.

    The layering for a KnitPackRepository is:

    Graph        |  HPSS    | Repository public layer |
    ===================================================
    Tuple based apis below, string based, and key based apis above
    ---------------------------------------------------
    VersionedFiles
      Provides .texts, .revisions etc
      This adapts the N-tuple keys to physical knit records which only have a
      single string identifier (for historical reasons), which in older formats
      was always the revision_id, and in the mapped code for packs is always
      the last element of key tuples.
    ---------------------------------------------------
    GraphIndex
      A separate GraphIndex is used for each of the
      texts/inventories/revisions/signatures contained within each individual
      pack file. The GraphIndex layer works in N-tuples and is unaware of any
      semantic value.
    ===================================================

    """

    # These attributes are inherited from the Repository base class. Setting
    # them to None ensures that if the constructor is changed to not initialize
    # them, or a subclass fails to call the constructor, that an error will
    # occur rather than the system working but generating incorrect data.
    _commit_builder_class: Type[VersionedFileCommitBuilder]
    _serializer: Serializer

    def __init__(self, _format, a_controldir, control_files, _commit_builder_class,
                 _serializer):
        MetaDirRepository.__init__(self, _format, a_controldir, control_files)
        self._commit_builder_class = _commit_builder_class
        self._serializer = _serializer
        self._reconcile_fixes_text_parents = True
        if self._format.supports_external_lookups:
            self._unstacked_provider = graph.CachingParentsProvider(
                self._make_parents_provider_unstacked())
        else:
            self._unstacked_provider = graph.CachingParentsProvider(self)
        self._unstacked_provider.disable_cache()

    def _all_revision_ids(self):
        """See Repository.all_revision_ids()."""
        with self.lock_read():
            return [key[0] for key in self.revisions.keys()]

    def _abort_write_group(self):
        self.revisions._index._key_dependencies.clear()
        self._pack_collection._abort_write_group()

    def _make_parents_provider(self):
        if not self._format.supports_external_lookups:
            return self._unstacked_provider
        return graph.StackedParentsProvider(_LazyListJoin(
            [self._unstacked_provider], self._fallback_repositories))

    def _refresh_data(self):
        if not self.is_locked():
            return
        self._pack_collection.reload_pack_names()
        self._unstacked_provider.disable_cache()
        self._unstacked_provider.enable_cache()

    def _start_write_group(self):
        self._pack_collection._start_write_group()

    def _commit_write_group(self):
        hint = self._pack_collection._commit_write_group()
        self.revisions._index._key_dependencies.clear()
        # The commit may have added keys that were previously cached as
        # missing, so reset the cache.
        self._unstacked_provider.disable_cache()
        self._unstacked_provider.enable_cache()
        return hint

    def suspend_write_group(self):
        # XXX check self._write_group is self.get_transaction()?
        tokens = self._pack_collection._suspend_write_group()
        self.revisions._index._key_dependencies.clear()
        self._write_group = None
        return tokens

    def _resume_write_group(self, tokens):
        self._start_write_group()
        try:
            self._pack_collection._resume_write_group(tokens)
        except errors.UnresumableWriteGroup:
            self._abort_write_group()
            raise
        for pack in self._pack_collection._resumed_packs:
            self.revisions._index.scan_unvalidated_index(pack.revision_index)

    def get_transaction(self):
        if self._write_lock_count:
            return self._transaction
        else:
            return self.control_files.get_transaction()

    def is_locked(self):
        return self._write_lock_count or self.control_files.is_locked()

    def is_write_locked(self):
        return self._write_lock_count

    def lock_write(self, token=None):
        """Lock the repository for writes.

        :return: A breezy.repository.RepositoryWriteLockResult.
        """
        locked = self.is_locked()
        if not self._write_lock_count and locked:
            raise errors.ReadOnlyError(self)
        self._write_lock_count += 1
        if self._write_lock_count == 1:
            self._transaction = transactions.WriteTransaction()
        if not locked:
            if 'relock' in debug.debug_flags and self._prev_lock == 'w':
                note('%r was write locked again', self)
            self._prev_lock = 'w'
            self._unstacked_provider.enable_cache()
            for repo in self._fallback_repositories:
                # Writes don't affect fallback repos
                repo.lock_read()
            self._refresh_data()
        return RepositoryWriteLockResult(self.unlock, None)

    def lock_read(self):
        """Lock the repository for reads.

        :return: A breezy.lock.LogicalLockResult.
        """
        locked = self.is_locked()
        if self._write_lock_count:
            self._write_lock_count += 1
        else:
            self.control_files.lock_read()
        if not locked:
            if 'relock' in debug.debug_flags and self._prev_lock == 'r':
                note('%r was read locked again', self)
            self._prev_lock = 'r'
            self._unstacked_provider.enable_cache()
            for repo in self._fallback_repositories:
                repo.lock_read()
            self._refresh_data()
        return LogicalLockResult(self.unlock)

    def leave_lock_in_place(self):
        # not supported - raise an error
        raise NotImplementedError(self.leave_lock_in_place)

    def dont_leave_lock_in_place(self):
        # not supported - raise an error
        raise NotImplementedError(self.dont_leave_lock_in_place)

    def pack(self, hint=None, clean_obsolete_packs=False):
        """Compress the data within the repository.

        This will pack all the data to a single pack. In future it may
        recompress deltas or do other such expensive operations.
        """
        with self.lock_write():
            self._pack_collection.pack(
                hint=hint, clean_obsolete_packs=clean_obsolete_packs)

    def reconcile(self, other=None, thorough=False):
        """Reconcile this repository."""
        from .reconcile import PackReconciler
        with self.lock_write():
            reconciler = PackReconciler(self, thorough=thorough)
            return reconciler.reconcile()

    def _reconcile_pack(self, collection, packs, extension, revs, pb):
        raise NotImplementedError(self._reconcile_pack)

    @only_raises(errors.LockNotHeld, errors.LockBroken)
    def unlock(self):
        if self._write_lock_count == 1 and self._write_group is not None:
            self.abort_write_group()
            self._unstacked_provider.disable_cache()
            self._transaction = None
            self._write_lock_count = 0
            raise errors.BzrError(
                'Must end write group before releasing write lock on %s'
                % self)
        if self._write_lock_count:
            self._write_lock_count -= 1
            if not self._write_lock_count:
                transaction = self._transaction
                self._transaction = None
                transaction.finish()
        else:
            self.control_files.unlock()

        if not self.is_locked():
            self._unstacked_provider.disable_cache()
            for repo in self._fallback_repositories:
                repo.unlock()


class RepositoryFormatPack(MetaDirVersionedFileRepositoryFormat):
    """Format logic for pack structured repositories.

    This repository format has:
     - a list of packs in pack-names
     - packs in packs/NAME.pack
     - indices in indices/NAME.{iix,six,tix,rix}
     - knit deltas in the packs, knit indices mapped to the indices.
     - thunk objects to support the knits programming API.
     - a format marker of its own
     - an optional 'shared-storage' flag
     - an optional 'no-working-trees' flag
     - a LockDir lock
    """

    # Set this attribute in derived classes to control the repository class
    # created by open and initialize.
    repository_class: Type[PackRepository]
    # Set this attribute in derived classes to control the
    # _commit_builder_class that the repository objects will have passed to
    # their constructor.
    _commit_builder_class: Type[VersionedFileCommitBuilder]
    # Set this attribute in derived clases to control the _serializer that the
    # repository objects will have passed to their constructor.
    _serializer: Serializer
    # Packs are not confused by ghosts.
    supports_ghosts: bool = True
    # External references are not supported in pack repositories yet.
    supports_external_lookups: bool = False
    # Most pack formats do not use chk lookups.
    supports_chks: bool = False
    # What index classes to use
    index_builder_class: Type[index.GraphIndexBuilder]
    index_class: Type[object]
    _fetch_uses_deltas: bool = True
    fast_deltas: bool = False
    supports_funky_characters: bool = True
    revision_graph_can_have_wrong_parents: bool = True

    def initialize(self, a_controldir, shared=False):
        """Create a pack based repository.

        :param a_controldir: bzrdir to contain the new repository; must already
            be initialized.
        :param shared: If true the repository will be initialized as a shared
                       repository.
        """
        mutter('creating repository in %s.', a_controldir.transport.base)
        dirs = ['indices', 'obsolete_packs', 'packs', 'upload']
        builder = self.index_builder_class()
        files = [('pack-names', builder.finish())]
        utf8_files = [('format', self.get_format_string())]

        self._upload_blank_content(
            a_controldir, dirs, files, utf8_files, shared)
        repository = self.open(a_controldir=a_controldir, _found=True)
        self._run_post_repo_init_hooks(repository, a_controldir, shared)
        return repository

    def open(self, a_controldir, _found=False, _override_transport=None):
        """See RepositoryFormat.open().

        :param _override_transport: INTERNAL USE ONLY. Allows opening the
                                    repository at a slightly different url
                                    than normal. I.e. during 'upgrade'.
        """
        if not _found:
            format = RepositoryFormatMetaDir.find_format(a_controldir)
        if _override_transport is not None:
            repo_transport = _override_transport
        else:
            repo_transport = a_controldir.get_repository_transport(None)
        control_files = lockable_files.LockableFiles(repo_transport,
                                                     'lock', lockdir.LockDir)
        return self.repository_class(_format=self,
                                     a_controldir=a_controldir,
                                     control_files=control_files,
                                     _commit_builder_class=self._commit_builder_class,
                                     _serializer=self._serializer)


class RetryPackOperations(RetryWithNewPacks):
    """Raised when we are packing and we find a missing file.

    Meant as a signaling exception, to tell the RepositoryPackCollection.pack
    code it should try again.
    """

    internal_error = True

    _fmt = ("Pack files have changed, reload and try pack again."
            " context: %(context)s %(orig_error)s")


class _DirectPackAccess:
    """Access to data in one or more packs with less translation."""

    def __init__(self, index_to_packs, reload_func=None, flush_func=None):
        """Create a _DirectPackAccess object.

        :param index_to_packs: A dict mapping index objects to the transport
            and file names for obtaining data.
        :param reload_func: A function to call if we determine that the pack
            files have moved and we need to reload our caches. See
            breezy.repo_fmt.pack_repo.AggregateIndex for more details.
        """
        self._container_writer = None
        self._write_index = None
        self._indices = index_to_packs
        self._reload_func = reload_func
        self._flush_func = flush_func

    def add_raw_record(self, key, size, raw_data):
        """Add raw knit bytes to a storage area.

        The data is spooled to the container writer in one bytes-record per
        raw data item.

        :param key: key of the data segment
        :param size: length of the data segment
        :param raw_data: A bytestring containing the data.
        :return: An opaque index memo For _DirectPackAccess the memo is
            (index, pos, length), where the index field is the write_index
            object supplied to the PackAccess object.
        """
        p_offset, p_length = self._container_writer.add_bytes_record(
            raw_data, size, [])
        return (self._write_index, p_offset, p_length)

    def add_raw_records(self, key_sizes, raw_data):
        """Add raw knit bytes to a storage area.

        The data is spooled to the container writer in one bytes-record per
        raw data item.

        :param sizes: An iterable of tuples containing the key and size of each
            raw data segment.
        :param raw_data: A bytestring containing the data.
        :return: A list of memos to retrieve the record later. Each memo is an
            opaque index memo. For _DirectPackAccess the memo is (index, pos,
            length), where the index field is the write_index object supplied
            to the PackAccess object.
        """
        raw_data = b''.join(raw_data)
        if not isinstance(raw_data, bytes):
            raise AssertionError(
                'data must be plain bytes was %s' % type(raw_data))
        result = []
        offset = 0
        for key, size in key_sizes:
            result.append(
                self.add_raw_record(key, size, [raw_data[offset:offset + size]]))
            offset += size
        return result

    def flush(self):
        """Flush pending writes on this access object.

        This will flush any buffered writes to a NewPack.
        """
        if self._flush_func is not None:
            self._flush_func()

    def get_raw_records(self, memos_for_retrieval):
        """Get the raw bytes for a records.

        :param memos_for_retrieval: An iterable containing the (index, pos,
            length) memo for retrieving the bytes. The Pack access method
            looks up the pack to use for a given record in its index_to_pack
            map.
        :return: An iterator over the bytes of the records.
        """
        # first pass, group into same-index requests
        request_lists = []
        current_index = None
        for (index, offset, length) in memos_for_retrieval:
            if current_index == index:
                current_list.append((offset, length))
            else:
                if current_index is not None:
                    request_lists.append((current_index, current_list))
                current_index = index
                current_list = [(offset, length)]
        # handle the last entry
        if current_index is not None:
            request_lists.append((current_index, current_list))
        for index, offsets in request_lists:
            try:
                transport, path = self._indices[index]
            except KeyError:
                # A KeyError here indicates that someone has triggered an index
                # reload, and this index has gone missing, we need to start
                # over.
                if self._reload_func is None:
                    # If we don't have a _reload_func there is nothing that can
                    # be done
                    raise
                raise RetryWithNewPacks(index,
                                        reload_occurred=True,
                                        exc_info=sys.exc_info())
            try:
                reader = pack.make_readv_reader(transport, path, offsets)
                for names, read_func in reader.iter_records():
                    yield read_func(None)
            except _mod_transport.NoSuchFile:
                # A NoSuchFile error indicates that a pack file has gone
                # missing on disk, we need to trigger a reload, and start over.
                if self._reload_func is None:
                    raise
                raise RetryWithNewPacks(transport.abspath(path),
                                        reload_occurred=False,
                                        exc_info=sys.exc_info())

    def set_writer(self, writer, index, transport_packname):
        """Set a writer to use for adding data."""
        if index is not None:
            self._indices[index] = transport_packname
        self._container_writer = writer
        self._write_index = index

    def reload_or_raise(self, retry_exc):
        """Try calling the reload function, or re-raise the original exception.

        This should be called after _DirectPackAccess raises a
        RetryWithNewPacks exception. This function will handle the common logic
        of determining when the error is fatal versus being temporary.
        It will also make sure that the original exception is raised, rather
        than the RetryWithNewPacks exception.

        If this function returns, then the calling function should retry
        whatever operation was being performed. Otherwise an exception will
        be raised.

        :param retry_exc: A RetryWithNewPacks exception.
        """
        is_error = False
        if self._reload_func is None:
            is_error = True
        elif not self._reload_func():
            # The reload claimed that nothing changed
            if not retry_exc.reload_occurred:
                # If there wasn't an earlier reload, then we really were
                # expecting to find changes. We didn't find them, so this is a
                # hard error
                is_error = True
        if is_error:
            # GZ 2017-03-27: No real reason this needs the original traceback.
            raise retry_exc.exc_info[1]



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import re
import sys
import textwrap
from io import StringIO

from .. import commands, export_pot, option, registry, tests


class TestEscape(tests.TestCase):

    def test_simple_escape(self):
        self.assertEqual(
            export_pot._escape('foobar'),
            'foobar')

        s = '''foo\nbar\r\tbaz\\"spam"'''
        e = '''foo\\nbar\\r\\tbaz\\\\\\"spam\\"'''
        self.assertEqual(export_pot._escape(s), e)

    def test_complex_escape(self):
        s = '''\\r \\\n'''
        e = '''\\\\r \\\\\\n'''
        self.assertEqual(export_pot._escape(s), e)


class TestNormalize(tests.TestCase):

    def test_single_line(self):
        s = 'foobar'
        e = '"foobar"'
        self.assertEqual(export_pot._normalize(s), e)

        s = 'foo"bar'
        e = '"foo\\"bar"'
        self.assertEqual(export_pot._normalize(s), e)

    def test_multi_lines(self):
        s = 'foo\nbar\n'
        e = '""\n"foo\\n"\n"bar\\n"'
        self.assertEqual(export_pot._normalize(s), e)

        s = '\nfoo\nbar\n'
        e = ('""\n'
             '"\\n"\n'
             '"foo\\n"\n'
             '"bar\\n"')
        self.assertEqual(export_pot._normalize(s), e)


class TestParseSource(tests.TestCase):
    """Check mappings to line numbers generated from python source"""

    def test_classes(self):
        src = '''
class Ancient:
    """Old style class"""

class Modern(object):
    """New style class"""
'''
        cls_lines, _ = export_pot._parse_source(src)
        self.assertEqual(cls_lines,
                         {"Ancient": 2, "Modern": 5})

    def test_classes_nested(self):
        src = '''
class Matroska(object):
    class Smaller(object):
        class Smallest(object):
            pass
'''
        cls_lines, _ = export_pot._parse_source(src)
        self.assertEqual(cls_lines,
                         {"Matroska": 2, "Smaller": 3, "Smallest": 4})

    def test_strings_docstrings(self):
        src = '''\
"""Module"""

def function():
    """Function"""

class Class(object):
    """Class"""

    def method(self):
        """Method"""
'''
        _, str_lines = export_pot._parse_source(src)
        self.assertEqual(str_lines,
                         {"Module": 1, "Function": 4, "Class": 7, "Method": 10})

    def test_strings_literals(self):
        src = '''\
s = "One"
t = (2, "Two")
f = dict(key="Three")
'''
        _, str_lines = export_pot._parse_source(src)
        self.assertEqual(str_lines,
                         {"One": 1, "Two": 2, "Three": 3})

    def test_strings_multiline(self):
        src = '''\
"""Start

End
"""
t = (
    "A"
    "B"
    "C"
    )
'''
        _, str_lines = export_pot._parse_source(src)
        self.assertEqual(str_lines,
                         {"Start\n\nEnd\n": 1, "ABC": 6})

    def test_strings_multiline_escapes(self):
        src = '''\
s = "Escaped\\n"
r = r"Raw\\n"
t = (
    "A\\n\\n"
    "B\\n\\n"
    "C\\n\\n"
    )
'''
        _, str_lines = export_pot._parse_source(src)
        if sys.version_info < (3, 8):
            self.expectFailure("Escaped newlines confuses the multiline handling",
                               self.assertNotEqual, str_lines,
                               {"Escaped\n": 0, "Raw\\n": 2, "A\n\nB\n\nC\n\n": -2})
        else:
            self.assertEqual(
                str_lines, {"Escaped\n": 1, "Raw\\n": 2, "A\n\nB\n\nC\n\n": 4})


class TestModuleContext(tests.TestCase):
    """Checks for source context tracking objects"""

    def check_context(self, context, path, lineno):
        self.assertEqual((context.path, context.lineno), (path, lineno))

    def test___init__(self):
        context = export_pot._ModuleContext("one.py")
        self.check_context(context, "one.py", 1)
        context = export_pot._ModuleContext("two.py", 5)
        self.check_context(context, "two.py", 5)

    def test_from_class(self):
        """New context returned with lineno updated from class"""
        path = "cls.py"

        class A:
            pass

        class B:
            pass
        cls_lines = {"A": 5, "B": 7}
        context = export_pot._ModuleContext(path, _source_info=(cls_lines, {}))
        contextA = context.from_class(A)
        self.check_context(contextA, path, 5)
        contextB1 = context.from_class(B)
        self.check_context(contextB1, path, 7)
        contextB2 = contextA.from_class(B)
        self.check_context(contextB2, path, 7)
        self.check_context(context, path, 1)
        self.assertEqual("", self.get_log())

    def test_from_class_missing(self):
        """When class has no lineno the old context details are returned"""
        path = "cls_missing.py"

        class A:
            pass

        class M:
            pass
        context = export_pot._ModuleContext(path, 3, ({"A": 15}, {}))
        contextA = context.from_class(A)
        contextM1 = context.from_class(M)
        self.check_context(contextM1, path, 3)
        contextM2 = contextA.from_class(M)
        self.check_context(contextM2, path, 15)
        self.assertContainsRe(self.get_log(), "Definition of <.*M'> not found")

    def test_from_string(self):
        """New context returned with lineno updated from string"""
        path = "str.py"
        str_lines = {"one": 14, "two": 42}
        context = export_pot._ModuleContext(path, _source_info=({}, str_lines))
        context1 = context.from_string("one")
        self.check_context(context1, path, 14)
        context2A = context.from_string("two")
        self.check_context(context2A, path, 42)
        context2B = context1.from_string("two")
        self.check_context(context2B, path, 42)
        self.check_context(context, path, 1)
        self.assertEqual("", self.get_log())

    def test_from_string_missing(self):
        """When string has no lineno the old context details are returned"""
        path = "str_missing.py"
        context = export_pot._ModuleContext(path, 4, ({}, {"line\n": 21}))
        context1 = context.from_string("line\n")
        context2A = context.from_string("not there")
        self.check_context(context2A, path, 4)
        context2B = context1.from_string("not there")
        self.check_context(context2B, path, 21)
        self.assertContainsRe(self.get_log(), "String b?'not there' not found")


class TestWriteOption(tests.TestCase):
    """Tests for writing texts extracted from options in pot format"""

    def pot_from_option(self, opt, context=None, note="test"):
        sio = StringIO()
        exporter = export_pot._PotExporter(sio)
        if context is None:
            context = export_pot._ModuleContext("nowhere", 0)
        export_pot._write_option(exporter, context, opt, note)
        return sio.getvalue()

    def test_option_without_help(self):
        opt = option.Option("helpless")
        self.assertEqual("", self.pot_from_option(opt))

    def test_option_with_help(self):
        opt = option.Option("helpful", help="Info.")
        self.assertContainsString(self.pot_from_option(opt), "\n"
                                  "# help of 'helpful' test\n"
                                  "msgid \"Info.\"\n")

    def test_option_hidden(self):
        opt = option.Option("hidden", help="Unseen.", hidden=True)
        self.assertEqual("", self.pot_from_option(opt))

    def test_option_context_missing(self):
        context = export_pot._ModuleContext("remote.py", 3)
        opt = option.Option("metaphor", help="Not a literal in the source.")
        self.assertContainsString(self.pot_from_option(opt, context),
                                  "#: remote.py:3\n"
                                  "# help of 'metaphor' test\n")

    def test_option_context_string(self):
        s = "Literally."
        context = export_pot._ModuleContext("local.py", 3, ({}, {s: 17}))
        opt = option.Option("example", help=s)
        self.assertContainsString(self.pot_from_option(opt, context),
                                  "#: local.py:17\n"
                                  "# help of 'example' test\n")

    def test_registry_option_title(self):
        opt = option.RegistryOption.from_kwargs("group", help="Pick one.",
                                                title="Choose!")
        pot = self.pot_from_option(opt)
        self.assertContainsString(pot, "\n"
                                  "# title of 'group' test\n"
                                  "msgid \"Choose!\"\n")
        self.assertContainsString(pot, "\n"
                                  "# help of 'group' test\n"
                                  "msgid \"Pick one.\"\n")

    def test_registry_option_title_context_missing(self):
        context = export_pot._ModuleContext("theory.py", 3)
        opt = option.RegistryOption.from_kwargs("abstract", title="Unfounded!")
        self.assertContainsString(self.pot_from_option(opt, context),
                                  "#: theory.py:3\n"
                                  "# title of 'abstract' test\n")

    def test_registry_option_title_context_string(self):
        s = "Grounded!"
        context = export_pot._ModuleContext("practice.py", 3, ({}, {s: 144}))
        opt = option.RegistryOption.from_kwargs("concrete", title=s)
        self.assertContainsString(self.pot_from_option(opt, context),
                                  "#: practice.py:144\n"
                                  "# title of 'concrete' test\n")

    def test_registry_option_value_switches(self):
        opt = option.RegistryOption.from_kwargs("switch", help="Flip one.",
                                                value_switches=True, enum_switch=False,
                                                red="Big.", green="Small.")
        pot = self.pot_from_option(opt)
        self.assertContainsString(pot, "\n"
                                  "# help of 'switch' test\n"
                                  "msgid \"Flip one.\"\n")
        self.assertContainsString(pot, "\n"
                                  "# help of 'switch=red' test\n"
                                  "msgid \"Big.\"\n")
        self.assertContainsString(pot, "\n"
                                  "# help of 'switch=green' test\n"
                                  "msgid \"Small.\"\n")

    def test_registry_option_value_switches_hidden(self):
        reg = registry.Registry()

        class Hider:
            hidden = True
        reg.register("new", 1, "Current.")
        reg.register("old", 0, "Legacy.", info=Hider())
        opt = option.RegistryOption("protocol", "Talking.", reg,
                                    value_switches=True, enum_switch=False)
        pot = self.pot_from_option(opt)
        self.assertContainsString(pot, "\n"
                                  "# help of 'protocol' test\n"
                                  "msgid \"Talking.\"\n")
        self.assertContainsString(pot, "\n"
                                  "# help of 'protocol=new' test\n"
                                  "msgid \"Current.\"\n")
        self.assertNotContainsString(pot, "'protocol=old'")


class TestPotExporter(tests.TestCase):
    """Test for logic specific to the _PotExporter class"""

    # This test duplicates test_duplicates below
    def test_duplicates(self):
        exporter = export_pot._PotExporter(StringIO())
        context = export_pot._ModuleContext("mod.py", 1)
        exporter.poentry_in_context(context, "Common line.")
        context.lineno = 3
        exporter.poentry_in_context(context, "Common line.")
        self.assertEqual(1, exporter.outf.getvalue().count("Common line."))

    def test_duplicates_included(self):
        exporter = export_pot._PotExporter(StringIO(), True)
        context = export_pot._ModuleContext("mod.py", 1)
        exporter.poentry_in_context(context, "Common line.")
        context.lineno = 3
        exporter.poentry_in_context(context, "Common line.")
        self.assertEqual(2, exporter.outf.getvalue().count("Common line."))


class PoEntryTestCase(tests.TestCase):

    def setUp(self):
        super().setUp()
        self.exporter = export_pot._PotExporter(StringIO())

    def check_output(self, expected):
        self.assertEqual(
            self.exporter.outf.getvalue(),
            textwrap.dedent(expected)
            )


class TestPoEntry(PoEntryTestCase):

    def test_simple(self):
        self.exporter.poentry('dummy', 1, "spam")
        self.exporter.poentry('dummy', 2, "ham", 'EGG')
        self.check_output('''\
                #: dummy:1
                msgid "spam"
                msgstr ""

                #: dummy:2
                # EGG
                msgid "ham"
                msgstr ""

                ''')

    def test_duplicate(self):
        self.exporter.poentry('dummy', 1, "spam")
        # This should be ignored.
        self.exporter.poentry('dummy', 2, "spam", 'EGG')

        self.check_output('''\
                #: dummy:1
                msgid "spam"
                msgstr ""\n
                ''')


class TestPoentryPerPergraph(PoEntryTestCase):

    def test_single(self):
        self.exporter.poentry_per_paragraph(
            'dummy',
            10,
            '''foo\nbar\nbaz\n'''
            )
        self.check_output('''\
                #: dummy:10
                msgid ""
                "foo\\n"
                "bar\\n"
                "baz\\n"
                msgstr ""\n
                ''')

    def test_multi(self):
        self.exporter.poentry_per_paragraph(
            'dummy',
            10,
            '''spam\nham\negg\n\nSPAM\nHAM\nEGG\n'''
            )
        self.check_output('''\
                #: dummy:10
                msgid ""
                "spam\\n"
                "ham\\n"
                "egg"
                msgstr ""

                #: dummy:14
                msgid ""
                "SPAM\\n"
                "HAM\\n"
                "EGG\\n"
                msgstr ""\n
                ''')


class TestExportCommandHelp(PoEntryTestCase):

    def test_command_help(self):

        class cmd_Demo(commands.Command):
            __doc__ = """A sample command.

            :Usage:
                bzr demo

            :Examples:
                Example 1::

                    cmd arg1

            Blah Blah Blah
            """

        export_pot._write_command_help(self.exporter, cmd_Demo())
        result = self.exporter.outf.getvalue()
        # We don't care about filename and lineno here.
        result = re.sub(r'(?m)^#: [^\n]+\n', '', result)

        self.assertEqualDiff(
            'msgid "A sample command."\n'
            'msgstr ""\n'
            '\n'                # :Usage: should not be translated.
            'msgid ""\n'
            '":Examples:\\n"\n'
            '"    Example 1::"\n'
            'msgstr ""\n'
            '\n'
            'msgid "        cmd arg1"\n'
            'msgstr ""\n'
            '\n'
            'msgid "Blah Blah Blah"\n'
            'msgstr ""\n'
            '\n',
            result
            )



# Copyright (C) 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import conflicts, tests
from breezy.bzr import conflicts as _mod_bzr_conflicts
from breezy.tests import KnownFailure, script
from breezy.tests.blackbox import test_conflicts


class TestResolve(script.TestCaseWithTransportAndScript):

    def setUp(self):
        super().setUp()
        test_conflicts.make_tree_with_conflicts(self, 'branch', 'other')

    def test_resolve_one_by_one(self):
        self.run_script("""\
$ cd branch
$ brz conflicts
Text conflict in my_other_file
Path conflict: mydir3 / mydir2
Text conflict in myfile
$ brz resolve myfile
2>1 conflict resolved, 2 remaining
$ brz resolve my_other_file
2>1 conflict resolved, 1 remaining
$ brz resolve mydir2
2>1 conflict resolved, 0 remaining
""")

    def test_resolve_all(self):
        self.run_script("""\
$ cd branch
$ brz resolve --all
2>3 conflicts resolved, 0 remaining
$ brz conflicts
""")

    def test_resolve_from_subdir(self):
        self.run_script("""\
$ mkdir branch/subdir
$ cd branch/subdir
$ brz resolve ../myfile
2>1 conflict resolved, 2 remaining
""")

    def test_resolve_via_directory_option(self):
        self.run_script("""\
$ brz resolve -d branch myfile
2>1 conflict resolved, 2 remaining
""")

    def test_resolve_all_via_directory_option(self):
        self.run_script("""\
$ brz resolve -d branch --all
2>3 conflicts resolved, 0 remaining
$ brz conflicts -d branch
""")

    def test_bug_842575_manual_rm(self):
        self.run_script("""\
$ brz init -q trunk
$ echo original > trunk/foo
$ brz add -q trunk/foo
$ brz commit -q -m first trunk
$ brz checkout -q trunk tree
$ brz rm -q trunk/foo
$ brz commit -q -m second trunk
$ echo modified > tree/foo
$ brz update tree
2>RM  foo => foo.THIS
2>Contents conflict in foo
2>1 conflicts encountered.
2>Updated to revision 2 of branch ...
$ rm tree/foo.BASE tree/foo.THIS
$ brz resolve --all -d tree
2>1 conflict resolved, 0 remaining
""")
        try:
            self.run_script("""\
$ brz status tree
""")
        except AssertionError:
            raise KnownFailure("bug #842575")

    def test_bug_842575_take_other(self):
        self.run_script("""\
$ brz init -q trunk
$ echo original > trunk/foo
$ brz add -q trunk/foo
$ brz commit -q -m first trunk
$ brz checkout -q --lightweight trunk tree
$ brz rm -q trunk/foo
$ brz ignore -d trunk foo
$ brz commit -q -m second trunk
$ echo modified > tree/foo
$ brz update tree
2>+N  .bzrignore
2>RM  foo => foo.THIS
2>Contents conflict in foo
2>1 conflicts encountered.
2>Updated to revision 2 of branch ...
$ brz resolve --take-other --all -d tree
2>1 conflict resolved, 0 remaining
""")
        try:
            self.run_script("""\
$ brz status tree
$ echo mustignore > tree/foo
$ brz status tree
""")
        except AssertionError:
            raise KnownFailure("bug 842575")


class TestBug788000(script.TestCaseWithTransportAndScript):

    def test_bug_788000(self):
        self.run_script('''\
$ brz init a
$ mkdir a/dir
$ echo foo > a/dir/file
$ brz add a/dir
$ cd a
$ brz commit -m one
$ cd ..
$ brz branch a b
$ echo bar > b/dir/file
$ cd a
$ rm -r dir
$ brz commit -m two
$ cd ../b
''',
                        null_output_matches_anything=True)

        self.run_script('''\
$ brz pull
Using saved parent location:...
Now on revision 2.
2>RM  dir/file => dir/file.THIS
2>Conflict: can't delete dir because it is not empty.  Not deleting.
2>Conflict because dir is not versioned, but has versioned children...
2>Contents conflict in dir/file
2>3 conflicts encountered.
''')
        self.run_script('''\
$ brz resolve --take-other
2>deleted dir/file.THIS
2>deleted dir
2>3 conflicts resolved, 0 remaining
''')


class TestResolveAuto(tests.TestCaseWithTransport):

    def test_auto_resolve(self):
        """Text conflicts can be resolved automatically"""
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/file',
                                   b'<<<<<<<\na\n=======\n>>>>>>>\n')])
        tree.add('file', ids=b'file_id')
        self.assertEqual(tree.kind('file'), 'file')
        file_conflict = _mod_bzr_conflicts.TextConflict('file', file_id=b'file_id')
        tree.set_conflicts([file_conflict])
        note = self.run_bzr('resolve', retcode=1, working_dir='tree')[1]
        self.assertContainsRe(note, '0 conflicts auto-resolved.')
        self.assertContainsRe(note,
                              'Remaining conflicts:\nText conflict in file')
        self.build_tree_contents([('tree/file', b'a\n')])
        note = self.run_bzr('resolve', working_dir='tree')[1]
        self.assertContainsRe(note, 'All conflicts resolved.')



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that all Tree's implement .annotate_iter()"""

from breezy.tests.per_tree import TestCaseWithTree


class TestAnnotate(TestCaseWithTree):

    def get_simple_tree(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([('tree/one', b'first\ncontent\n')])
        tree.add(['one'])
        rev_1 = tree.commit('one')
        self.build_tree_contents([('tree/one', b'second\ncontent\n')])
        rev_2 = tree.commit('two')
        return self._convert_tree(tree), [rev_1, rev_2]

    def get_tree_with_ghost(self):
        tree = self.make_branch_and_tree('tree')
        if not tree.branch.repository._format.supports_ghosts:
            self.skipTest('repository format does not support ghosts')
        self.build_tree_contents([('tree/one', b'first\ncontent\n')])
        tree.add(['one'])
        rev_1 = tree.commit('one')
        tree.set_parent_ids([rev_1, b'ghost-one'])
        self.build_tree_contents([('tree/one', b'second\ncontent\n')])
        rev_2 = tree.commit('two')
        return self._convert_tree(tree), [rev_1, rev_2]

    def test_annotate_simple(self):
        tree, revids = self.get_simple_tree()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([(revids[1], b'second\n'), (revids[0], b'content\n')],
                         list(tree.annotate_iter('one')))

    def test_annotate_with_ghost(self):
        tree, revids = self.get_tree_with_ghost()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([(revids[1], b'second\n'), (revids[0], b'content\n')],
                         list(tree.annotate_iter('one')))



# Copyright (C) 2006, 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""TreeBuilder helper class.

TreeBuilders are used to build trees of various shapes or properties. This
can be extremely useful in testing for instance.
"""

from . import errors


class AlreadyBuilding(errors.BzrError):

    _fmt = "The tree builder is already building a tree."


class NotBuilding(errors.BzrError):

    _fmt = "Not currently building a tree."


class TreeBuilder:
    """A TreeBuilder allows the creation of specific content in one tree at a
    time.
    """

    def __init__(self):
        """Construct a TreeBuilder."""
        self._tree = None
        self._root_done = False

    def build(self, recipe):
        """Build recipe into the current tree.

        :param recipe: A sequence of paths. For each path, the corresponding
            path in the current tree is created and added. If the path ends in
            '/' then a directory is added, otherwise a regular file is added.
        """
        self._ensure_building()
        if not self._root_done:
            self._tree.add('', 'directory', ids=b'root-id')
            self._root_done = True
        for name in recipe:
            if name.endswith('/'):
                self._tree.mkdir(name[:-1])
            else:
                end = b'\n'
                content = b"contents of %s%s" % (name.encode('utf-8'), end)
                self._tree.add(name, 'file')
                self._tree.put_file_bytes_non_atomic(name, content)

    def _ensure_building(self):
        """Raise NotBuilding if there is no current tree being built."""
        if self._tree is None:
            raise NotBuilding

    def finish_tree(self):
        """Finish building the current tree."""
        self._ensure_building()
        tree = self._tree
        self._tree = None
        tree.unlock()

    def start_tree(self, tree):
        """Start building on tree.

        :param tree: A tree to start building on. It must provide the
            MutableTree interface.
        """
        if self._tree is not None:
            raise AlreadyBuilding
        self._tree = tree
        self._tree.lock_tree_write()



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test read_bundle works properly across various transports."""

from io import BytesIO

import breezy.bzr.bzrdir
import breezy.mergeable
import breezy.transport
import breezy.urlutils

from ... import errors, tests
from ...tests.per_transport import transport_test_permutations
from ...tests.scenarios import load_tests_apply_scenarios
from ...tests.test_transport import TestTransportImplementation
from ..bundle.serializer import write_bundle

load_tests = load_tests_apply_scenarios


def create_bundle_file(test_case):
    test_case.build_tree(['tree/', 'tree/a', 'tree/subdir/'])

    format = breezy.bzr.bzrdir.BzrDirFormat.get_default_format()

    bzrdir = format.initialize('tree')
    repo = bzrdir.create_repository()
    branch = repo.controldir.create_branch()
    wt = branch.controldir.create_workingtree()

    wt.add(['a', 'subdir/'])
    wt.commit('new project', rev_id=b'commit-1')

    out = BytesIO()
    write_bundle(wt.branch.repository, wt.get_parent_ids()[0], b'null:', out)
    out.seek(0)
    return out, wt


class TestReadMergeableBundleFromURL(TestTransportImplementation):
    """Test that read_bundle works properly across multiple transports"""

    scenarios = transport_test_permutations()

    def setUp(self):
        super().setUp()
        self.bundle_name = 'test_bundle'
        # read_mergeable_from_url will invoke get_transport which may *not*
        # respect self._transport (i.e. returns a transport that is different
        # from the one we want to test, so we must inject a correct transport
        # into possible_transports first).
        self.possible_transports = [self.get_transport(self.bundle_name)]
        self.overrideEnv('BRZ_NO_SMART_VFS', None)
        self.create_test_bundle()

    def read_mergeable_from_url(self, url):
        return breezy.mergeable.read_mergeable_from_url(
            url, possible_transports=self.possible_transports)

    def get_url(self, relpath=''):
        return breezy.urlutils.join(self._server.get_url(), relpath)

    def create_test_bundle(self):
        out, wt = create_bundle_file(self)
        if self.get_transport().is_readonly():
            self.build_tree_contents([(self.bundle_name, out.getvalue())])
        else:
            self.get_transport().put_file(self.bundle_name, out)
            self.log('Put to: %s', self.get_url(self.bundle_name))
        return wt

    def test_read_mergeable_from_url(self):
        info = self.read_mergeable_from_url(
            str(self.get_url(self.bundle_name)))
        revision = info.real_revisions[-1]
        self.assertEqual(b'commit-1', revision.revision_id)

    def test_read_fail(self):
        # Trying to read from a directory, or non-bundle file
        # should fail with NotABundle
        self.assertRaises(errors.NotABundle,
                          self.read_mergeable_from_url, self.get_url('tree'))
        self.assertRaises(errors.NotABundle,
                          self.read_mergeable_from_url, self.get_url('tree/a'))

    def test_read_mergeable_respects_possible_transports(self):
        if not isinstance(self.get_transport(self.bundle_name),
                          breezy.transport.ConnectedTransport):
            # There is no point testing transport reuse for not connected
            # transports (the test will fail even).
            raise tests.TestSkipped(
                'Need a ConnectedTransport to test transport reuse')
        url = str(self.get_url(self.bundle_name))
        self.read_mergeable_from_url(url)
        self.assertEqual(1, len(self.possible_transports))



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import errno
import os
import subprocess
import sys
import tempfile
from typing import Type

import breezy

from . import config as _mod_config
from . import email_message, errors, msgeditor, osutils, registry, urlutils


class MailClientNotFound(errors.BzrError):

    _fmt = "Unable to find mail client with the following names:"\
        " %(mail_command_list_string)s"

    def __init__(self, mail_command_list):
        mail_command_list_string = ', '.join(mail_command_list)
        errors.BzrError.__init__(
            self, mail_command_list=mail_command_list,
            mail_command_list_string=mail_command_list_string)


class NoMessageSupplied(errors.BzrError):

    _fmt = "No message supplied."


class NoMailAddressSpecified(errors.BzrError):

    _fmt = "No mail-to address (--mail-to) or output (-o) specified."


class MailClient:
    """A mail client that can send messages with attachements."""

    def __init__(self, config):
        self.config = config

    def compose(self, prompt, to, subject, attachment, mime_subtype,
                extension, basename=None, body=None):
        """Compose (and possibly send) an email message

        Must be implemented by subclasses.

        :param prompt: A message to tell the user what to do.  Supported by
            the Editor client, but ignored by others
        :param to: The address to send the message to
        :param subject: The contents of the subject line
        :param attachment: An email attachment, as a bytestring
        :param mime_subtype: The attachment is assumed to be a subtype of
            Text.  This allows the precise subtype to be specified, e.g.
            "plain", "x-patch", etc.
        :param extension: The file extension associated with the attachment
            type, e.g. ".patch"
        :param basename: The name to use for the attachment, e.g.
            "send-nick-3252"
        """
        raise NotImplementedError

    def compose_merge_request(self, to, subject, directive, basename=None,
                              body=None):
        """Compose (and possibly send) a merge request

        :param to: The address to send the request to
        :param subject: The subject line to use for the request
        :param directive: A merge directive representing the merge request, as
            a bytestring.
        :param basename: The name to use for the attachment, e.g.
            "send-nick-3252"
        """
        prompt = self._get_merge_prompt("Please describe these changes:", to,
                                        subject, directive)
        self.compose(prompt, to, subject, directive,
                     'x-patch', '.patch', basename, body)

    def _get_merge_prompt(self, prompt, to, subject, attachment):
        """Generate a prompt string.  Overridden by Editor.

        :param prompt: A string suggesting what user should do
        :param to: The address the mail will be sent to
        :param subject: The subject line of the mail
        :param attachment: The attachment that will be used
        """
        return ''


mail_client_registry = registry.Registry[str, Type[MailClient]]()


class Editor(MailClient):
    __doc__ = """DIY mail client that uses commit message editor"""

    supports_body = True

    def _get_merge_prompt(self, prompt, to, subject, attachment):
        """See MailClient._get_merge_prompt"""
        return ("%s\n\n"
                "To: %s\n"
                "Subject: %s\n\n"
                "%s" % (prompt, to, subject,
                         attachment.decode('utf-8', 'replace')))

    def compose(self, prompt, to, subject, attachment, mime_subtype,
                extension, basename=None, body=None):
        """See MailClient.compose"""
        if not to:
            raise NoMailAddressSpecified()
        body = msgeditor.edit_commit_message(prompt, start_message=body)
        if body == '':
            raise NoMessageSupplied()
        email_message.EmailMessage.send(self.config,
                                        self.config.get('email'),
                                        to,
                                        subject,
                                        body,
                                        attachment,
                                        attachment_mime_subtype=mime_subtype)


mail_client_registry.register('editor', Editor,
                              help=Editor.__doc__)


class BodyExternalMailClient(MailClient):

    supports_body = True

    def _get_client_commands(self):
        """Provide a list of commands that may invoke the mail client"""
        if sys.platform == 'win32':
            import win32utils
            return [win32utils.get_app_path(i) for i in self._client_commands]
        else:
            return self._client_commands

    def compose(self, prompt, to, subject, attachment, mime_subtype,
                extension, basename=None, body=None):
        """See MailClient.compose.

        Writes the attachment to a temporary file, invokes _compose.
        """
        if basename is None:
            basename = 'attachment'
        pathname = tempfile.mkdtemp(prefix='bzr-mail-')
        attach_path = osutils.pathjoin(pathname, basename + extension)
        with open(attach_path, 'wb') as outfile:
            outfile.write(attachment)
        if body is not None:
            kwargs = {'body': body}
        else:
            kwargs = {}
        self._compose(prompt, to, subject, attach_path, mime_subtype,
                      extension, **kwargs)

    def _compose(self, prompt, to, subject, attach_path, mime_subtype,
                 extension, body=None, from_=None):
        """Invoke a mail client as a commandline process.

        Overridden by MAPIClient.
        :param to: The address to send the mail to
        :param subject: The subject line for the mail
        :param pathname: The path to the attachment
        :param mime_subtype: The attachment is assumed to have a major type of
            "text", but the precise subtype can be specified here
        :param extension: A file extension (including period) associated with
            the attachment type.
        :param body: Optional body text.
        :param from_: Optional From: header.
        """
        for name in self._get_client_commands():
            cmdline = [self._encode_path(name, 'executable')]
            if body is not None:
                kwargs = {'body': body}
            else:
                kwargs = {}
            if from_ is not None:
                kwargs['from_'] = from_
            cmdline.extend(self._get_compose_commandline(to, subject,
                                                         attach_path,
                                                         **kwargs))
            try:
                subprocess.call(cmdline)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
            else:
                break
        else:
            raise MailClientNotFound(self._client_commands)

    def _get_compose_commandline(self, to, subject, attach_path, body):
        """Determine the commandline to use for composing a message

        Implemented by various subclasses
        :param to: The address to send the mail to
        :param subject: The subject line for the mail
        :param attach_path: The path to the attachment
        """
        raise NotImplementedError

    def _encode_safe(self, u):
        """Encode possible unicode string argument to 8-bit string
        in user_encoding. Unencodable characters will be replaced
        with '?'.

        :param  u:  possible unicode string.
        :return:    encoded string if u is unicode, u itself otherwise.
        """
        return u

    def _encode_path(self, path, kind):
        """Encode unicode path in user encoding.

        :param  path:   possible unicode path.
        :param  kind:   path kind ('executable' or 'attachment').
        :return:        encoded path if path is unicode,
                        path itself otherwise.
        :raise:         UnableEncodePath.
        """
        return path


class ExternalMailClient(BodyExternalMailClient):
    __doc__ = """An external mail client."""

    supports_body = False


class Evolution(BodyExternalMailClient):
    __doc__ = """Evolution mail client."""

    _client_commands = ['evolution']

    def _get_compose_commandline(self, to, subject, attach_path, body=None):
        """See ExternalMailClient._get_compose_commandline"""
        message_options = {}
        if subject is not None:
            message_options['subject'] = subject
        if attach_path is not None:
            message_options['attach'] = attach_path
        if body is not None:
            message_options['body'] = body
        options_list = ['{}={}'.format(k, urlutils.escape(v)) for (k, v) in
                        sorted(message_options.items())]
        return ['mailto:{}?{}'.format(self._encode_safe(to or ''),
                                  '&'.join(options_list))]


mail_client_registry.register('evolution', Evolution,
                              help=Evolution.__doc__)


class Mutt(BodyExternalMailClient):
    __doc__ = """Mutt mail client."""

    _client_commands = ['mutt']

    def _get_compose_commandline(self, to, subject, attach_path, body=None):
        """See ExternalMailClient._get_compose_commandline"""
        message_options = []
        if subject is not None:
            message_options.extend(
                ['-s', self._encode_safe(subject)])
        if attach_path is not None:
            message_options.extend(
                ['-a', self._encode_path(attach_path, 'attachment')])
        if body is not None:
            # Store the temp file object in self, so that it does not get
            # garbage collected and delete the file before mutt can read it.
            self._temp_file = tempfile.NamedTemporaryFile(
                prefix="mutt-body-", suffix=".txt", mode="w+")
            self._temp_file.write(body)
            self._temp_file.flush()
            message_options.extend(['-i', self._temp_file.name])
        if to is not None:
            message_options.extend(['--', self._encode_safe(to)])
        return message_options


mail_client_registry.register('mutt', Mutt,
                              help=Mutt.__doc__)


class Thunderbird(BodyExternalMailClient):
    __doc__ = """Mozilla Thunderbird (or Icedove)

    Note that Thunderbird 1.5 is buggy and does not support setting
    "to" simultaneously with including a attachment.

    There is a workaround if no attachment is present, but we always need to
    send attachments.
    """

    _client_commands = [
        'thunderbird', 'mozilla-thunderbird', 'icedove',
        '/Applications/Mozilla/Thunderbird.app/Contents/MacOS/thunderbird-bin',
        '/Applications/Thunderbird.app/Contents/MacOS/thunderbird-bin']

    def _get_compose_commandline(self, to, subject, attach_path, body=None):
        """See ExternalMailClient._get_compose_commandline"""
        message_options = {}
        if to is not None:
            message_options['to'] = self._encode_safe(to)
        if subject is not None:
            message_options['subject'] = self._encode_safe(subject)
        if attach_path is not None:
            message_options['attachment'] = urlutils.local_path_to_url(
                attach_path)
        if body is not None:
            options_list = ['body=%s' %
                            urlutils.quote(self._encode_safe(body))]
        else:
            options_list = []
        options_list.extend(["{}='{}'".format(k, v) for k, v in
                             sorted(message_options.items())])
        return ['-compose', ','.join(options_list)]


mail_client_registry.register('thunderbird', Thunderbird,
                              help=Thunderbird.__doc__)


class KMail(ExternalMailClient):
    __doc__ = """KDE mail client."""

    _client_commands = ['kmail']

    def _get_compose_commandline(self, to, subject, attach_path):
        """See ExternalMailClient._get_compose_commandline"""
        message_options = []
        if subject is not None:
            message_options.extend(['-s', self._encode_safe(subject)])
        if attach_path is not None:
            message_options.extend(
                ['--attach', self._encode_path(attach_path, 'attachment')])
        if to is not None:
            message_options.extend([self._encode_safe(to)])
        return message_options


mail_client_registry.register('kmail', KMail,
                              help=KMail.__doc__)


class Claws(ExternalMailClient):
    __doc__ = """Claws mail client."""

    supports_body = True

    _client_commands = ['claws-mail']

    def _get_compose_commandline(self, to, subject, attach_path, body=None,
                                 from_=None):
        """See ExternalMailClient._get_compose_commandline"""
        compose_url = []
        if from_ is not None:
            compose_url.append('from=' + urlutils.quote(from_))
        if subject is not None:
            # Don't use urlutils.quote_plus because Claws doesn't seem
            # to recognise spaces encoded as "+".
            compose_url.append(
                'subject=' + urlutils.quote(self._encode_safe(subject)))
        if body is not None:
            compose_url.append(
                'body=' + urlutils.quote(self._encode_safe(body)))
        # to must be supplied for the claws-mail --compose syntax to work.
        if to is None:
            raise NoMailAddressSpecified()
        compose_url = 'mailto:{}?{}'.format(
            self._encode_safe(to), '&'.join(compose_url))
        # Collect command-line options.
        message_options = ['--compose', compose_url]
        if attach_path is not None:
            message_options.extend(
                ['--attach', self._encode_path(attach_path, 'attachment')])
        return message_options

    def _compose(self, prompt, to, subject, attach_path, mime_subtype,
                 extension, body=None, from_=None):
        """See ExternalMailClient._compose"""
        if from_ is None:
            from_ = self.config.get('email')
        super()._compose(prompt, to, subject, attach_path,
                                    mime_subtype, extension, body, from_)


mail_client_registry.register('claws', Claws,
                              help=Claws.__doc__)


class XDGEmail(BodyExternalMailClient):
    __doc__ = """xdg-email attempts to invoke the preferred mail client"""

    _client_commands = ['xdg-email']

    def _get_compose_commandline(self, to, subject, attach_path, body=None):
        """See ExternalMailClient._get_compose_commandline"""
        if not to:
            raise NoMailAddressSpecified()
        commandline = [self._encode_safe(to)]
        if subject is not None:
            commandline.extend(['--subject', self._encode_safe(subject)])
        if attach_path is not None:
            commandline.extend(['--attach',
                                self._encode_path(attach_path, 'attachment')])
        if body is not None:
            commandline.extend(['--body', self._encode_safe(body)])
        return commandline


mail_client_registry.register('xdg-email', XDGEmail,
                              help=XDGEmail.__doc__)


class EmacsMail(ExternalMailClient):
    __doc__ = """Call emacsclient to have a mail buffer.

    This only work for emacs >= 22.1 due to recent -e/--eval support.

    The good news is that this implementation will work with all mail
    agents registered against ``mail-user-agent``. So there is no need
    to instantiate ExternalMailClient for each and every GNU Emacs
    MUA.

    Users just have to ensure that ``mail-user-agent`` is set according
    to their tastes.
    """

    _client_commands = ['emacsclient']

    def __init__(self, config):
        super().__init__(config)
        self.elisp_tmp_file = None

    def _prepare_send_function(self):
        """Write our wrapper function into a temporary file.

        This temporary file will be loaded at runtime in
        _get_compose_commandline function.

        This function does not remove the file.  That's a wanted
        behaviour since _get_compose_commandline won't run the send
        mail function directly but return the eligible command line.
        Removing our temporary file here would prevent our sendmail
        function to work.  (The file is deleted by some elisp code
        after being read by Emacs.)
        """

        _defun = br"""(defun bzr-add-mime-att (file)
  "Attach FILE to a mail buffer as a MIME attachment."
  (let ((agent mail-user-agent))
    (if (and file (file-exists-p file))
        (cond
         ((eq agent 'sendmail-user-agent)
          (progn
            (mail-text)
            (newline)
            (if (functionp 'etach-attach)
              (etach-attach file)
              (mail-attach-file file))))
         ((or (eq agent 'message-user-agent)
              (eq agent 'gnus-user-agent)
              (eq agent 'mh-e-user-agent))
          (progn
            (mml-attach-file file "text/x-patch" "BZR merge" "inline")))
         ((eq agent 'mew-user-agent)
          (progn
            (mew-draft-prepare-attachments)
            (mew-attach-link file (file-name-nondirectory file))
            (let* ((nums (mew-syntax-nums))
                   (syntax (mew-syntax-get-entry mew-encode-syntax nums)))
              (mew-syntax-set-cd syntax "BZR merge")
              (mew-encode-syntax-print mew-encode-syntax))
            (mew-header-goto-body)))
         (t
          (message "Unhandled MUA, report it on bazaar@lists.canonical.com")))
      (error "File %s does not exist." file))))
"""

        fd, temp_file = tempfile.mkstemp(prefix="emacs-bzr-send-",
                                         suffix=".el")
        try:
            os.write(fd, _defun)
        finally:
            os.close(fd)  # Just close the handle but do not remove the file.
        return temp_file

    def _get_compose_commandline(self, to, subject, attach_path):
        commandline = ["--eval"]

        _to = "nil"
        _subject = "nil"

        if to is not None:
            _to = ("\"%s\"" % self._encode_safe(to).replace('"', '\\"'))
        if subject is not None:
            _subject = ("\"%s\"" %
                        self._encode_safe(subject).replace('"', '\\"'))

        # Funcall the default mail composition function
        # This will work with any mail mode including default mail-mode
        # User must tweak mail-user-agent variable to tell what function
        # will be called inside compose-mail.
        mail_cmd = "(compose-mail {} {})".format(_to, _subject)
        commandline.append(mail_cmd)

        # Try to attach a MIME attachment using our wrapper function
        if attach_path is not None:
            # Do not create a file if there is no attachment
            elisp = self._prepare_send_function()
            self.elisp_tmp_file = elisp
            lmmform = '(load "%s")' % elisp
            mmform = '(bzr-add-mime-att "%s")' % \
                self._encode_path(attach_path, 'attachment')
            rmform = '(delete-file "%s")' % elisp
            commandline.append(lmmform)
            commandline.append(mmform)
            commandline.append(rmform)

        return commandline


mail_client_registry.register('emacsclient', EmacsMail,
                              help=EmacsMail.__doc__)


class MAPIClient(BodyExternalMailClient):
    __doc__ = """Default Windows mail client launched using MAPI."""

    def _compose(self, prompt, to, subject, attach_path, mime_subtype,
                 extension, body=None):
        """See ExternalMailClient._compose.

        This implementation uses MAPI via the simplemapi ctypes wrapper
        """
        from .util import simplemapi
        try:
            simplemapi.SendMail(to or '', subject or '', body or '',
                                attach_path)
        except simplemapi.MAPIError as e:
            if e.code != simplemapi.MAPI_USER_ABORT:
                raise MailClientNotFound(['MAPI supported mail client'
                                          ' (error %d)' % (e.code,)])


mail_client_registry.register('mapi', MAPIClient,
                              help=MAPIClient.__doc__)


class MailApp(BodyExternalMailClient):
    __doc__ = """Use MacOS X's Mail.app for sending email messages.

    Although it would be nice to use appscript, it's not installed
    with the shipped Python installations.  We instead build an
    AppleScript and invoke the script using osascript(1).  We don't
    use the _encode_safe() routines as it's not clear what encoding
    osascript expects the script to be in.
    """

    _client_commands = ['osascript']

    def _get_compose_commandline(self, to, subject, attach_path, body=None,
                                 from_=None):
        """See ExternalMailClient._get_compose_commandline"""

        fd, self.temp_file = tempfile.mkstemp(prefix="bzr-send-",
                                              suffix=".scpt")
        try:
            os.write(fd, 'tell application "Mail"\n')
            os.write(fd, 'set newMessage to make new outgoing message\n')
            os.write(fd, 'tell newMessage\n')
            if to is not None:
                os.write(fd, 'make new to recipient with properties'
                         ' {address:"%s"}\n' % to)
            if from_ is not None:
                # though from_ doesn't actually seem to be used
                os.write(fd, 'set sender to "%s"\n'
                         % from_.replace('"', '\\"'))
            if subject is not None:
                os.write(fd, 'set subject to "%s"\n'
                         % subject.replace('"', '\\"'))
            if body is not None:
                # FIXME: would be nice to prepend the body to the
                # existing content (e.g., preserve signature), but
                # can't seem to figure out the right applescript
                # incantation.
                os.write(fd, 'set content to "%s\\n\n"\n' %
                         body.replace('"', '\\"').replace('\n', '\\n'))

            if attach_path is not None:
                # FIXME: would be nice to first append a newline to
                # ensure the attachment is on a new paragraph, but
                # can't seem to figure out the right applescript
                # incantation.
                os.write(fd, 'tell content to make new attachment'
                         ' with properties {file name:"%s"}'
                         ' at after the last paragraph\n'
                         % self._encode_path(attach_path, 'attachment'))
            os.write(fd, 'set visible to true\n')
            os.write(fd, 'end tell\n')
            os.write(fd, 'end tell\n')
        finally:
            os.close(fd)  # Just close the handle but do not remove the file.
        return [self.temp_file]


mail_client_registry.register('mail.app', MailApp,
                              help=MailApp.__doc__)


class DefaultMail(MailClient):
    __doc__ = """Default mail handling.  Tries XDGEmail (or MAPIClient on Windows),
    falls back to Editor"""

    supports_body = True

    def _mail_client(self):
        """Determine the preferred mail client for this platform"""
        if osutils.supports_mapi():
            return MAPIClient(self.config)
        else:
            return XDGEmail(self.config)

    def compose(self, prompt, to, subject, attachment, mime_subtype,
                extension, basename=None, body=None):
        """See MailClient.compose"""
        try:
            return self._mail_client().compose(prompt, to, subject,
                                               attachment, mime_subtype,
                                               extension, basename, body)
        except MailClientNotFound:
            return Editor(self.config).compose(
                prompt, to, subject, attachment, mime_subtype, extension, body)

    def compose_merge_request(self, to, subject, directive, basename=None,
                              body=None):
        """See MailClient.compose_merge_request"""
        try:
            return self._mail_client().compose_merge_request(
                to, subject, directive, basename=basename, body=body)
        except MailClientNotFound:
            return Editor(self.config).compose_merge_request(
                to, subject, directive, basename=basename, body=body)


mail_client_registry.register('default', DefaultMail,
                              help=DefaultMail.__doc__)
mail_client_registry.default_key = 'default'

opt_mail_client = _mod_config.RegistryOption(
    'mail_client', mail_client_registry, help='E-mail client to use.',
    invalid='error')



# Copyright (C) 2007-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz version."""

import os
import sys

import breezy
from breezy import osutils, trace
from breezy.tests import (TestCase, TestCaseInTempDir, TestSkipped,
                          probe_unicode_in_user_encoding)


class TestVersion(TestCase):

    def test_main_version(self):
        """Check output from version command and master option is reasonable"""
        # output is intentionally passed through to stdout so that we
        # can see the version being tested
        self.permit_source_tree_branch_repo()
        output = self.run_bzr('version')[0]
        self.log('brz version output:')
        self.log(output)
        self.assertTrue(output.startswith('Breezy (brz) '))
        self.assertNotEqual(output.index('Canonical'), -1)
        # make sure --version is consistent
        tmp_output = self.run_bzr('--version')[0]
        self.assertEqual(output, tmp_output)

    def test_version(self):
        self.permit_source_tree_branch_repo()
        out = self.run_bzr("version")[0]
        self.assertTrue(len(out) > 0)
        self.assertEqualDiff(out.splitlines()[0],
                             "Breezy (brz) %s" % breezy.__version__)
        self.assertContainsRe(out, r"(?m)^  Python interpreter:")
        self.assertContainsRe(out, r"(?m)^  Python standard library:")
        self.assertContainsRe(out, r"(?m)^  breezy:")
        self.assertContainsRe(out, r"(?m)^  Breezy configuration:")
        self.assertContainsRe(out, r'(?m)^  Breezy log file:.*[\\/]breezy[\\/]brz\.log')

    def test_version_short(self):
        self.permit_source_tree_branch_repo()
        out = self.run_bzr(["version", "--short"])[0]
        self.assertEqualDiff(out, breezy.version_string + '\n')


class TestVersionUnicodeOutput(TestCaseInTempDir):

    def _check(self, args):
        self.permit_source_tree_branch_repo()
        # Even though trace._brz_log_filename variable
        # is used only to keep actual log filename
        # and changing this variable in selftest
        # don't change main brz.log location,
        # and therefore pretty safe,
        # but we run these tests in separate temp dir
        # with relative unicoded path
        old_trace_file = trace._brz_log_filename
        trace._brz_log_filename = '\u1234/.brz.log'
        try:
            out = self.run_bzr(args)[0]
        finally:
            trace._brz_log_filename = old_trace_file
        self.assertTrue(len(out) > 0)
        self.assertContainsRe(out, r'(?m)^  Breezy log file:.*brz\.log')

    def test_command(self):
        self._check("version")

    def test_flag(self):
        self._check("--version")

    def test_unicode_bzr_home(self):
        uni_val, str_val = probe_unicode_in_user_encoding()
        if uni_val is None:
            raise TestSkipped('Cannot find a unicode character that works in'
                              ' encoding %s' % (osutils.get_user_encoding(),))

        self.overrideEnv('BRZ_HOME', uni_val)
        self.permit_source_tree_branch_repo()
        out = self.run_bzr_raw("version")[0]
        self.assertTrue(len(out) > 0)
        self.assertContainsRe(out, br"(?m)^  Breezy configuration: " + str_val)


class TestVersionBzrLogLocation(TestCaseInTempDir):

    def default_log(self):
        return os.path.join(os.environ['BRZ_HOME'], 'breezy', 'brz.log')

    def test_simple(self):
        brz_log = 'my.brz.log'
        self.overrideEnv('BRZ_LOG', brz_log)
        self.assertPathDoesNotExist([self.default_log(), brz_log])
        out = self.run_brz_subprocess('version')[0]
        self.assertTrue(len(out) > 0)
        self.assertContainsRe(
            out, br"(?m)^  Breezy log file: " + brz_log.encode('ascii'))
        self.assertPathExists(brz_log)
        self.assertPathDoesNotExist(self.default_log())

    def test_dev_null(self):
        # This test uses a subprocess to cause the log opening logic to
        # execute. It would be better to just execute that logic directly.
        if sys.platform == 'win32':
            brz_log = 'NUL'
        else:
            brz_log = '/dev/null'
        self.overrideEnv('BRZ_LOG', brz_log)
        self.assertPathDoesNotExist(self.default_log())
        out = self.run_brz_subprocess('version')[0]
        self.assertTrue(len(out) > 0)
        self.assertContainsRe(
            out, br"(?m)^  Breezy log file: " + brz_log.encode('ascii'))
        self.assertPathDoesNotExist(self.default_log())

    def test_unicode_brz_log(self):
        uni_val = "\xa7"
        enc = osutils.get_user_encoding()
        try:
            str_val = uni_val.encode(enc)
        except UnicodeEncodeError:
            self.skipTest(
                "Test string {!r} unrepresentable in user encoding {}".format(
                    uni_val, enc))
        brz_log = os.path.join(self.test_base_dir, uni_val)
        self.overrideEnv("BRZ_LOG", brz_log)
        out, err = self.run_brz_subprocess("version")
        uni_out = out.decode(enc)
        self.assertContainsRe(uni_out, "(?m)^  Breezy log file: .*/\xa7$")
